"""RaESL plotting utility functions."""

from collections import defaultdict
from copy import deepcopy
from typing import Dict, Generator, List, Set, Tuple

from ragraph.edge import Edge
from ragraph.graph import Graph
from ragraph.node import Node
from ragraph.plot.utils import get_axis_sequence


def get_component_function_dicts(
    graph: Graph,
) -> Tuple[Dict[str, List[Node]], Dict[str, List[Node]]]:
    """Get two dictionaries from component node names to lists of transformation and
    goal function nodes.

    Arguments
       graph: Instantiated ESL graph.

    Returns
       Dictionaries from component node names to transformation and goal function nodes.
    """
    trans = defaultdict(list)
    goals = defaultdict(list)
    for f in graph.get_nodes_by_kind(kind="function_spec"):
        if f.annotations.esl_info["sub_kind"] == "transformation":
            trans[f.annotations.esl_info["body"]["active"]].append(f)
        elif f.annotations.esl_info["sub_kind"] == "goal":
            goals[f.annotations.esl_info["body"]["active"]].append(f)

    return trans, goals


def crawl_descendants(node: Node, max_depth: int) -> Tuple[List[Node], List[Edge]]:
    """Walk down the decomposition tree of a node, add its children and create
    composition dependencies accordingly.

    Arguments:
        node: The parent node to unfold.
        max_depth: Max depth to unfold nodes to (absolute depth w.r.t. root).

    Returns:
        List of (additionally) crawled nodes and "composition_dependency" edges.

    Note:
        Since we explicitly need to figure out which (partial) hierarchies to display
        using Graphviz, we recreate those parts of the hierarchy using explicit
        "composition dependency" :obj:`Edge` objects in the preprocessing step using
        :obj:`ragraph.graph.GraphView` objects.
    """
    nodes, edges = [], []

    for child in node.children:
        nodes.append(child)
        edges.append(Edge(node, child, kind="composition_dependency"))

        if child.children and child.depth < max_depth:
            cnodes, cedges = crawl_descendants(child, max_depth)
            nodes.extend(cnodes)
            edges.extend(cedges)

    return nodes, edges


def get_neighbors(
    graph: Graph, root: Node, node: Node, node_kinds: Set[str], degree: int = 1
) -> Set[Node]:
    """Get neighbors of a node within the graph.

    Arguments:
        graph: Instantiated ESL graph.
        root: The root node for which the nodes must be sought up to the given degree.
        node: The node for which the direct neighbors must be collected.
        node_kinds: The kinds of the neighbor nodes that must be collected.
        degree: The degree up to which neighbors must be collected (neighbors of
        neighbors). Defaults to 1.

    Returns
        Set of neighbors.
    """

    neighbors = set(
        [
            neighbor
            for neighbor in list(graph.targets_of(node)) + list(graph.sources_of(node))
            if (
                neighbor.kind in node_kinds
                and (
                    neighbor.depth == root.depth
                    or (neighbor.depth < root.depth and not neighbor.children)
                )
            )
        ]
    )

    if degree == 1:
        # Stop searching further.
        return neighbors

    for neighbor in neighbors:
        neighbors = neighbors.union(
            get_neighbors(graph, root, neighbor, node_kinds, degree=degree - 1)
        )

    return neighbors


def get_paths_between_all(
    sources: List[Node], targets: List[Node], edges: List[Edge]
) -> List[List[str]]:
    """Compute paths between nodes. Based on depth first search.

    Arguments:
         sources: List of starting nodes of paths.
         targets: Set of ending nodes of path.
         edges: List of edges between nodes.

    Yields:
        List of lists of node names.
    """
    paths = []

    ed: Dict[str, Dict[str, Edge]] = defaultdict(dict)
    for e in edges:
        ed[e.source][e.target] = e

    for source in sources:
        paths.extend(get_paths(source, ed, set(targets), visited=[source.name]))

    return paths


def get_paths(
    source: Node,
    edges: Dict[Node, Dict[Node, Edge]],
    targets: Set[Node],
    visited: List[Node] = [],
) -> List[List[str]]:
    """Collection all paths (list of node names) between the source node and the set of
    target nodes.

    Arguments:
        source: The source node where all paths should start.
        edges: Dictionary of Node to Node to Edge. Contains the edges to be
            considered when searching for paths.
        targets: Set of node where the paths should end.
        visited: List of nodes already visited. Required to prevent running in cycles.

    Returns:
        List of lists of node names.
    """
    paths: List[List[str]] = []
    steps = edges.get(source)
    if not steps:
        return paths

    for target in steps:
        if target.name in visited:
            # node has already been visited. Loop entered.
            continue
        elif target in targets:
            # A target has been reached
            visited.append(target.name)
            paths.append(visited)
        else:
            new_path = deepcopy(visited)
            new_path.append(target.name)
            paths.extend(get_paths(target, edges, targets, visited=new_path))

    return paths


def select_components_for_function_path(
    graph: Graph,
    start_points: List[Tuple[Node, List[Node]]],
    end_points: List[Tuple[Node, List[Node]]],
    levels: int,
) -> List[Node]:
    """Select components that belong to a certain level within the decomposition
    structure to draw a functional path.

    Arguments:
        graph: Instantiated ESL graph.
        start_points: List of tuples that contain the component node and list of
            function nodes that serve as a starting point of the function chains.
        end_points: List of Tuples that contain the component node and list of functions
            that nodes server.
        levels: Number of levels to decompose intermediate components into.
            This number is relative to the maximum of the depth of the start
        node and end node.

    Note:
       Components that are a descendant of the start or end component node of the
       function path are excluded.
    """
    start_comps = [p[0] for p in start_points]
    end_comps = [p[0] for p in end_points]
    level = max([n.depth for n in start_comps + end_comps]) + levels - 1
    start_descendants = get_all_descendants(start_comps)
    end_descendants = get_all_descendants(end_comps)
    components = []

    for c in graph.get_nodes_by_kind(kind="component"):
        if c in start_descendants or c in end_descendants or c.depth > level:
            continue
        elif c.depth == level or (c.depth < level and not c.children):
            components.append(c)

    return components


def add_level(graph: Graph, sliced: Graph, parents: Set[Node]):
    """Add a level to the sliced graph based on the hierarchy in the source graph.

    Arguments:
        graph: Instantiated ESL graph.
        sliced: Sliced data graph.

    Note:
        Creates deepcopies of child nodes to preserve the original data structure.
    """
    comp_names = set([c.name for c in sliced.get_nodes_by_kind(kind="component")])

    for p in parents:
        children = comp_names.intersection(set([child.name for child in p.children]))
        if children:
            n = deepcopy(p)
            n.children = [sliced[child] for child in children]
            sliced.add_node(n)


def rebuild_hierarchical_structure(graph: Graph, sliced: Graph):
    """Rebuilding the hierarchical structure within the sliced graph.

    Arguments:
        graph: Instantiated ESL graph.
        sliced: Sliced data graph.

    Note:
        Creates deepcopies of child nodes to preserve the original data structure.
    """
    components = sliced.get_nodes_by_kind(kind="component")
    depth = max([graph[c.name].depth for c in components]) if components else 0
    parents = set(
        [
            graph[c.name].parent
            for c in components
            if graph[c.name].parent is not None and graph[c.name].parent.depth == depth - 1
        ]
    )

    while parents:
        add_level(graph, sliced, parents=parents)
        depth -= 1
        comps = sliced.get_nodes_by_kind(kind="component")
        parents = set(
            [
                graph[c.name].parent
                for c in comps
                if graph[c.name].parent is not None and graph[c.name].parent.depth == depth - 1
            ]
        )


def get_all_descendants(nodes: List[Node]) -> Set[Node]:
    """Get all descendants from a list of provided nodes.

    Arguments:
        nodes: List of nodes for which the descendants must be found.

    Returns:
        Set of nodes containing all descendants of the provided nodes.
    """
    all_descendants: Set[Node] = set()
    for n in nodes:
        all_descendants.union(set(n.descendants))

    return all_descendants


def check_comp_func_pairs(points: List[Tuple[Node, List[Node]]]):
    """Check of for all points the provided components functions pairs are consistent.

    Arguments:
        points: List of points for which the consistency must be checked.

    Raises:
        ValueError: If the active component of the given functions does not match.
    """
    for point in points:
        comp = point[0]
        functions = point[1]
        for f in functions:
            if comp.name != f.annotations.esl_info["body"]["active"]:
                msg = "Active component of '{}' is not equal to start component '{}'."
                raise ValueError(msg.format(comp.name, f.annotations.esl_info["body"]["active"]))


def check_tree_disjunction(nodes: List[Node]):
    """Check if list of nodes are not part of each others descendants.

    Raises:
        ValueError: If node is in the descendants of another node that was provided.
    """
    decendants = get_all_descendants(nodes)
    for n in nodes:
        if n in decendants:
            raise ValueError("Node {} is a descendant of another provided node.".format(n.name))


def check_start_and_end_points(
    start_points: List[Tuple[Node, List[Node]]],
    end_points: List[Tuple[Node, List[Node]]],
):
    """Check start and end points for consistency.

    Arguments:
        start_points: List of tuples that contain the component node and list of
            function nodes that serve as starting points of the function chains.
        end_points: List of tuples that contain the component node and list of
            functions nodes that serve as end points of the function chains.

    Raises:
        ValueError: If the component-function pairs do not have a corresponding active
            component.
        ValueError: If any of the components are a descendant of another component.
    """
    points = start_points + end_points
    check_comp_func_pairs(points)
    check_tree_disjunction([p[0] for p in points])


def get_function_tree(graph: Graph, node: Node, levels: int) -> Tuple[Set[Node], Set[Edge]]:
    """Walk down the traceability tree of a node.

    Arguments:
        graph: Instantiated ESL graph.
        node: Node to start the walk from.
        levels: Number of levels to continue walking.

    Returns:
        Tuple of the set of all nodes and the set of all edges that describe the
        traceability tree.

    """
    edges = set()
    functions = set()
    functions.add(node)

    if levels > 1:
        edges.update(
            set([e for e in graph.edges_from(node) if e.kind == "traceability_dependency"])
        )

        function_targets = set([e.target for e in edges])
        functions.update(function_targets)
        for f in function_targets:
            if f.annotations.esl_info["sub_kind"] == "goal":
                continue

            fs, es = get_function_tree(graph, f, levels - 1)

            edges.update(es)
            functions.update(fs)

    return functions, edges


def migrate_edges_between_variables(graph: Graph, lead_components: List[Node]) -> List[Edge]:
    """Migrate edges between variables up into the decomposition tree.

    Arguments:
        graph: Instantiated ESL graph.
        lead_components: The components that are displayed.

    Returns:
        List of edges to are created for the specific view.

    Note:
        The ESL compiler only adds functional dependencies between variables based on
        the transformation specs of leaf components. As such, when one displays
        components at higher hierarchical levels gaps appear in the variable dependency
        structure. Hence, this method is used to fill those gaps by adding additional
        edges.
    """
    edges = []
    for c in lead_components:
        if not c.children:
            continue
        # Component is a collapsed component so get transformation specs and add
        # dependencies between in and output variables.
        tfs = [
            n
            for n in graph.targets_of(c)
            if n.kind == "function_spec" and n.annotations.esl_info["sub_kind"] == "transformation"
        ]
        for t in tfs:
            for vi in t.annotations.esl_info["body"]["input_variables"]:
                for vj in t.annotations.esl_info["body"]["output_variables"]:
                    edges.append(
                        Edge(
                            source=graph[vi],
                            target=graph[vj],
                            labels=[graph[vi].annotations.esl_info["type_ref"]],
                            kind="functional_dependency",
                            annotations=dict(esl_info=dict(reason=dict(transformations=[t.name]))),
                        )
                    )
    return edges


def has_mapping_dependency(graph: Graph, node: Node, nodes: List[Node]) -> bool:
    """Check if a node is mapped to any node in a list of nodes.

    Arguments:
        graph: Instantiated ESL graph.
        node: The node for which the existence of mapping dependencies must be checked.
        nodes: List of nodes which are to be considered for checking the existence
            of a mapping dependency.

    Returns:
        Bool that indicates if a mapping dependency exists.
    """
    return any(
        True for e in graph.edges_between_all(nodes, [node]) if e.kind == "mapping_dependency"
    ) or any(True for e in graph.edges_between_all([node], nodes) if e.kind == "mapping_dependency")


def filter_nodes(graph: Graph, lead_components: List[Node], node_kinds: List[str]) -> List[Node]:
    """Filter nodes for displaying a multi-domain-matrix.

    Arguments:
        graph: Instantiated ESL graph.
        lead_components: List of nodes of kind 'component' that are leading in
           filter the other nodes.
        node_kinds: List of node kinds to be included.

    Returns:
        List of filtered nodes.

    Note:
        Node of kind 'component' are always leading in filtering the nodes to
            ensure the consistency within the network of shown dependencies. Since
            the component hierarchy forms the central structure of an ESL specification.
    """
    nodes = []
    for kind in node_kinds:
        if kind == "component":
            nodes.extend(lead_components)
        else:
            domain_nodes = []
            for n in graph.get_nodes_by_kind(kind=kind):
                if has_mapping_dependency(graph, n, lead_components + nodes):
                    domain_nodes.append(n)

            nodes.extend(get_axis_sequence(nodes=domain_nodes, kinds=[kind]))

    return nodes


def get_up_to_depth(roots: List[Node], depth: int) -> Generator[Node, None, None]:
    """Get nodes up to a certain depth with bus nodes at the start of child lists.

    Arguments:
        roots: List of nodes to walk down from.
        depth: Depth up to which nodes must be returned.

    Returns:
        Nodes up to the provided list.
    """
    for node in roots:
        if node.is_leaf or node.depth == depth:
            yield node
            continue

        children = sorted(node.children, key=lambda n: n.is_bus, reverse=True)

        if node.depth == depth - 1:
            yield from children
        else:
            yield from get_up_to_depth(children, depth)



"""Code for collecting and type checking of transformations."""
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from raesl.compile.ast import components
from raesl.compile.typechecking import utils
from raesl.compile.typechecking.expr_checker import ExprChecker
from raesl.compile.typechecking.goal_transform_base import GoalTransformBaseBuilder

if TYPE_CHECKING:
    from raesl.compile.ast.components import ComponentDefinition, SubClause
    from raesl.compile.ast.specification import Specification
    from raesl.compile.scanner import Token
    from raesl.compile.typechecking.compdef_builder import CompDefChildBuilders


class CompDefTransformBuilder(GoalTransformBaseBuilder):
    """Collect transformations of a component from the parser, check them, and add them
    to the component definition.

    Arguments:
        comp_child_builders: Storage of child builders for a component definition.

    Attributes:
        transforms: Collected transformations.
        transform_kind: Last found kind of transformation kind, either 'requirement' or
            'constraint'.
    """

    def __init__(self, comp_child_builders: "CompDefChildBuilders"):
        super(CompDefTransformBuilder, self).__init__(comp_child_builders.diag_store)
        self.transforms: List[components.Transformation] = []
        self.comp_child_builders = comp_child_builders
        self.transform_kind: Optional[str] = None

    def new_transform_header(self, transform_kind: "Token"):
        """New transform section line found.

        Arguments:
            transform_kind: Kind of transformations that will follow.
        """
        if transform_kind.tok_type == "TRANSFORM_REQUIREMENT_KW":
            self.transform_kind = components.REQUIREMENT
        else:
            assert transform_kind.tok_type == "TRANSFORM_CONSTRAINT_KW"
            self.transform_kind = components.CONSTRAINT

    def add_transform(self, transform: components.Transformation):
        """A new transformation has been found, add it to the collection.

        Arguments:
            transform: Transformation to add.
        """
        assert self.transform_kind is not None
        transform.transform_kind = self.transform_kind
        self.transforms.append(transform)

    def add_transform_subclause(self, sub_clause: "SubClause"):
        """Add a found subclause that belongs to the last transformation."""
        self.transforms[-1].sub_clauses.append(sub_clause)

    def finish_comp(self, comp_def: "ComponentDefinition", spec: "Specification"):
        """Check the found transformations, and add them to the component.

        Arguments:
            comp_def: Component definition to extend with the found relation instances.
                Also a source of available variables, parameters, and variable groups.
            spec: Specification being constructed. Source for types and relation
                definitions processed previously.
        """
        vps = utils.construct_var_param_map(comp_def)
        vpps = utils.construct_verb_prepos_combis(spec)

        expr_checker = ExprChecker(vps, self.diag_store)

        # Verify all transformations in the component.
        good_transforms = []  # Transformations to add to the component.
        elements_by_label: Dict[str, List[Any]] = self.comp_child_builders.elements_by_label
        for trans in self.transforms:
            is_good = True
            assert trans.transform_kind is not None
            self.check_form(
                "transformation",
                trans.transform_kind,
                trans.doesaux_tok,
                trans.sub_clauses,
            )

            # Store transformation on its label for duplicate label detection.
            elements_by_label[trans.label_tok.tok_text].append(trans)

            if not self.verify_flows(trans.in_flows, vps):
                is_good = False
            if not self.verify_flows(trans.out_flows, vps):
                is_good = False

            self.verify_verb_prepos(trans.verb_tok, trans.prepos_tok, vpps)

            # Verify subclauses.
            for sub in trans.sub_clauses:
                if not expr_checker.check_expr(sub.expr):
                    is_good = False

            if is_good:
                good_transforms.append(trans)

        comp_def.transforms = good_transforms



"""Line matcher state machines needs."""
from typing import TYPE_CHECKING

from raesl.compile.machine_files import typing

if TYPE_CHECKING:
    from raesl.compile.typechecking.ast_builder import AstBuilder


_NEED_HEADER_SPEC = """
needs_header:
    start initial;
    start -> s1 [NEED_KW] tag=need;

    end accept=need_header;
    s1 -> end [NL_TK];
    s1 -> end [EOF_TK];
"""


def _process_need_header(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    builder.notify_new_section(tags["need"][0], False)


_NEED_LINE_SPEC = """
needs_line:
    start initial;
    start -> s1 [NAME] tag=label;
    s1 -> s2 [COLON_TK];
    s2 -> s3 [DOTTEDNAME] tag=subject;
    s3 -> s4 [NONSPACE] tag=description;

    s4 -> s4 [NONSPACE] tag=description;

    end accept=need_line;
    s4 -> end [NL_TK];
    s4 -> end [EOF_TK];
"""


def _process_need(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    label_tok = tags["label"][0]
    subject_tok = tags["subject"][0]
    description = " ".join(tag.tok_text for tag in tags["description"])

    current_comp = builder.compdef_builder.current_component
    assert current_comp is not None
    current_comp.add_need(label_tok, subject_tok, description)


MACHINES: typing.MachineTripletList = [
    ("NEED_HEADER_MACHINE", _NEED_HEADER_SPEC, _process_need_header),
    ("NEED_LINE_MACHINE", _NEED_LINE_SPEC, _process_need),
]



"""ESL compiler.

Compiles ESL documents and workspaces, meaning:

- Parsing lines
- Typechecking
- Building an AST
- Instantiating components, variables and requirements.
- Deriving dependencies from these to build an output graph (network).
"""
from pathlib import Path
from typing import Optional, Union

import raesl.compile.cli
from ragraph.graph import Graph


class EslCompilationError(Exception):
    """Error during ESL compilation."""


def to_graph(
    *paths: Union[str, Path],
    output: Optional[Union[str, Path]] = None,
    force: bool = False,
) -> Graph:
    """Convert ESL file(s) into a :obj:`ragraph.graph.Graph`.

    Arguments:
        paths: Paths to resolve into ESL files. May be any number of files and
            directories to scan.
        output: Optional output file (JSON) to write the graph to.
        force: Whether to overwrite the output file or raise an error if it the file
            already exists.

    Returns:
        Instantiated graph.
    """

    diag, spec, graph = raesl.compile.cli.run(*paths, output=output, force=force)

    if graph is None:
        errors = "\n".join(str(d) for d in diag.diagnostics)
        raise EslCompilationError(
            f"Could not compile the specification into a Graph object:\n{errors}"
        )

    return graph



"""Methods for casting ESL AST into ragraph.node.Node objects."""
import re
from collections import defaultdict
from typing import Any, Dict, List, Optional, Union

from raesl import logger
from raesl.compile import diagnostics
from raesl.compile.ast import exprs, types
from raesl.compile.ast.components import (
    BehaviorFunction,
    ComponentDefinition,
    ComponentInstance,
    Design,
    Flow,
    Goal,
    InstanceArgument,
    Need,
    RelationInstance,
    Transformation,
    VarParam,
)
from raesl.compile.ast.nodes import (
    CompoundVarNode,
    ElementaryVarNode,
    GroupNode,
    VarNode,
)
from raesl.compile.ast.specification import Specification
from raesl.compile.ast.types import ElementaryType, TypeDef
from raesl.compile.instantiating.graph_data import InstNode
from ragraph.node import Node


class NodeStore:
    """Node storage with multiple catagories for quicker access to specific subsets."""

    categories = [
        "nodes",
        "types",
        "components",
        "variables",
        "needs",
        "goals",
        "transforms",
        "designs",
        "behaviors",
        "relations",
    ]

    def __init__(self):
        for cat in self.categories:
            setattr(self, cat, dict())

    def clear(self):
        """Clear all node categories."""
        for cat in self.categories:
            getattr(self, cat).clear()

    def add(self, node: Node, *args):
        """Add node to :obj:`self.nodes` and any other specified categories in args."""
        name = node.name
        self.nodes[name] = node
        for m in args:
            getattr(self, m)[name] = node
        logger.debug(f"Added ragraph.node.Node '{node.name}' to nodes and {args}.")


class NodeFactory:
    """Node factory. Creates :obj:`Node` objects from a :obj:`Specification`."""

    def __init__(
        self,
        diag_store: diagnostics.DiagnosticStore,
        spec: Optional[Specification] = None,
    ):
        self.diag_store = diag_store
        self.spec = spec
        self.node_store = NodeStore()
        self.type_inst_map = {}

    def _add(self, node: Optional[Node], *args):
        """Proxy for :obj:`NodeStore.add`."""
        self.node_store.add(node, *args)

    def make_nodes(self, spec: Optional[Specification] = None) -> Dict[str, Node]:
        """Instantiate AST and create :obj:`Node` objects accordingly."""
        # Clear maps.
        self.node_store.clear()

        # Set provided spec or try and keep the current one.
        self.spec = self.spec if spec is None else spec
        if self.spec is None:
            return dict()

        # Create types first.
        for t in self.spec.types.values():
            self.type_inst_map[t.type] = t
            tn = make_type_node(t)
            if tn is not None:
                self._add(tn, "types")

        # Create world node and instantiate it (recursively).
        self._add(
            Node(
                name="world",
                kind="component",
                annotations=dict(
                    esl_info=dict(definition_name="world", property_variables=[], comments=[])
                ),
            ),
            "components",
        )
        self._instantiate_component(spec.world, {}, "world")

        post_process_comments(self.node_store.nodes.values())

        return self.node_store.nodes

    def _instantiate_component(
        self,
        comp_def: ComponentDefinition,
        inst_map: Dict[ElementaryVarNode, InstNode],
        inst_name: str,
    ):
        """Instantiate a component definition.

        Arguments
            comp_def: Component definition to instantiate.
            inst_map: Mapping of parameter VarNodes of the component definition
                to InstNodes of the parent. Is not used by the parent afterwards.
            inst_name: Dotted name of the component instance.
        """
        # Keep a separate list of the component variables for property counting checks
        local_varinst_list: List[InstNode] = []
        for var in comp_def.variables:
            varname = [inst_name, var.name_tok.tok_text]  # Dotted (sub) var name.
            var_map = make_variable_instmap(var, varname, var.node)
            local_varinst_list.extend(inst_map.values())
            inst_map.update(var_map)

        # Dump mapping of elementary variable nodes to instance nodes.
        logger.debug("Variable node map of instance {}".format(inst_name))
        logger.debug("\n".join(str(k) + ":" + str(v) for k, v in inst_map.items()))
        logger.debug("")

        # Creating ragraph.node.Node object for all new instance nodes.
        for v in inst_map.values():
            self._add(make_variable_node(v, self.type_inst_map), "variables")

        # Construct child component instances, recursing down to the leaf components.
        for comp_inst in comp_def.component_instances:
            assert comp_inst.compdef is not None
            assert len(comp_inst.arguments) == len(comp_inst.compdef.parameters)

            child_inst_map: Dict[ElementaryVarNode, InstNode] = {}
            for arg, param in zip(comp_inst.arguments, comp_inst.compdef.parameters):
                assert arg.argnode is not None
                result = make_parameter_instmap(param, param.node, arg.argnode, inst_map)
                child_inst_map.update(result)

            # Full dotted name.
            child_inst_name = f"{inst_name}.{comp_inst.inst_name_tok.tok_text}"

            # Add component as a node.
            self._add(
                make_component_node(c=comp_inst, inst_name=child_inst_name, params=child_inst_map),
                "components",
            )
            self._instantiate_component(comp_inst.compdef, child_inst_map, child_inst_name)

            # Set parent of child
            self.node_store.components[child_inst_name].parent = self.node_store.components[
                inst_name
            ]

        # Verify ownership of the variable InstNodes.
        for inst_node in local_varinst_list:
            inst_node.check_owner(self.diag_store)

        # Instantiate need, goal, transform, design, behavior and relation specs.
        for n in comp_def.needs:
            self._add(make_need_node(n, inst_name), "needs")

        for g in comp_def.goals:
            self._add(make_goal_node(g, inst_name, inst_map), "goals")

        for t in comp_def.transforms:
            self._add(make_transform_node(t, inst_name, inst_map), "transforms")

        for d in comp_def.designs:
            self._add(make_design_node(d, inst_name, inst_map), "designs")

        for b in comp_def.behaviors:
            self._add(make_behavior_node(b, inst_name, inst_map), "behaviors")

        for r in comp_def.relations:
            self._add(make_relation_node(r, inst_name, inst_map), "relations")


def make_variable_instmap(
    var: VarParam, varname: List[str], node: VarNode
) -> Dict[ElementaryVarNode, InstNode]:
    """Construct instance nodes for the provided variable.

    Arguments:
        var: Variable represented by the node.
        varname: Collected prefixes of the dotted name so far.
        node: Node to associate with one or more inst nodes.

    Returns:
        Map of the elementary variable nodes to their associated instance nodes.
    """
    assert var.is_variable

    if isinstance(node, ElementaryVarNode):
        instnode = InstNode(".".join(varname), node)
        instnode.add_comment(node.get_comment())
        return {node: instnode}
    else:
        assert isinstance(node, CompoundVarNode)
        varname.append("")  # Will be overwritten in upcoming loop
        inst_map = {}
        for cn in node.child_nodes:
            varname[-1] = cn.name_tok.tok_text
            inst_map.update(make_variable_instmap(var, varname, cn))
        del varname[-1]
        return inst_map


def make_parameter_instmap(
    param: VarParam,
    param_node: VarNode,
    arg_node: VarNode,
    parent_inst_map: Dict[ElementaryVarNode, InstNode],
) -> Dict[ElementaryVarNode, InstNode]:
    """Construct an inst node map for a parameter of a child component instance.

    Arguments:
        param: Parameter definition in the child component definition.
        param_node: VarNode within the parameter in the child component
            definition. Note these are like variables in the child.
        arg_node: Node in the parent that must match with param_node.
            These nodes exist in the parent component, and may contain part of a
            variable group.
        parent_inst_map: Variable instance map of the parent. As component
            instantiation is recursive, this may also include nodes from the
            grand-parent or higher. Should not be modified.

    Returns:
        Instance node map for a parameter.
    """
    if isinstance(param_node, ElementaryVarNode):
        # Elementary parameter node. At this point, the arg_node should also be
        # elementary, and exist as instance node representing part of a variable
        # in the # parent.
        # Link that instance node to the param node in the child.
        assert isinstance(arg_node, ElementaryVarNode)

        instnode = parent_inst_map.get(arg_node)
        assert instnode is not None

        instnode.add_param(param)

        node = param.resolve_element(param.name_tok.tok_text)
        if isinstance(node, ElementaryVarNode):
            instnode.add_comment(node.get_comment())
        elif isinstance(node, CompoundVarNode):
            child = node.resolve_node(instnode.name.split(".")[-1])
            if child:
                instnode.add_comment(child.get_comment())
        return {param_node: instnode}
    else:
        child_map = {}
        assert isinstance(param_node, CompoundVarNode)
        if isinstance(arg_node, CompoundVarNode):
            assert len(param_node.child_nodes) == len(arg_node.child_nodes)
            for param_child, arg_child in zip(param_node.child_nodes, arg_node.child_nodes):
                child_map.update(
                    make_parameter_instmap(param, param_child, arg_child, parent_inst_map)
                )
        else:
            assert isinstance(arg_node, GroupNode)
            assert len(param_node.child_nodes) == len(arg_node.child_nodes)
            for param_child, arg_child in zip(param_node.child_nodes, arg_node.child_nodes):
                child_map.update(
                    make_parameter_instmap(param, param_child, arg_child, parent_inst_map)
                )

        return child_map


def make_type_node(tdef: types.TypeDef) -> Node:
    """Node creation for a type definition.

    Arguments:
      tdef: Type definition for which a node most be created.

    Returns:
      Node of kind `variabel_type`
    """
    a = _make_type_annotations(tdef=tdef)
    if not a:
        return
    return Node(name=tdef.name.tok_text, kind="variable_type", annotations=a)


def _make_type_annotations(tdef: types.TypeDef) -> Dict[str, Dict[str, Any]]:
    """Creating dictionary to annotate `variable_type` nodes.

    Arguments:
      tdef: The type definition for which an annotation dictionary most be created.

    Returns:
      See YAML output spec /components/schemas/VariableTypeAnnotation
    """
    if not isinstance(tdef.type, types.ElementaryType):
        return

    units = [tok.tok_text for tok in tdef.type.units]
    domain = None
    if not tdef.type.intervals:
        return dict(esl_info=dict(units=units, domain=domain))

    domain = []

    for interval in tdef.type.intervals:
        d = dict(
            lowerbound=dict(value=None, unit=None),
            upperbound=dict(value=None, unit=None),
        )
        if interval[0]:
            if interval[0].value:
                d["lowerbound"]["value"] = interval[0].value.tok_text
            if interval[0].unit:
                d["lowerbound"]["unit"] = interval[0].unit.tok_text
        if interval[1]:
            if interval[1].value:
                d["upperbound"]["value"] = interval[1].value.tok_text
            if interval[1].unit:
                d["upperbound"]["unit"] = interval[1].unit.tok_text
        domain.append(d)

    return dict(esl_info=dict(units=units, domain=domain))


def make_variable_node(v: InstNode, type_inst_map: Dict[ElementaryType, TypeDef]) -> Node:
    """Node creation for a variable.

    Arguments:
        v: The variable for which a nodes must be created.

    Returns:
        Node of kind "variable".
    """
    a = _make_variable_annotations(v=v, type_inst_map=type_inst_map)
    return Node(name=v.name, kind="variable", annotations=a)


def _make_variable_annotations(
    v: InstNode,
    type_inst_map: Dict[ElementaryType, TypeDef],
    comments: Optional[List[str]] = [],
) -> Dict[str, Dict[str, Any]]:
    """Creating dictionary for variable annotations.

    Arguments:
        v: The variable for which the annotation dict must be created.
        comments: List of comments attached to the nodes.

    Returns:
        See YAML output spec /components/schemas/VariableAnnotation
    """
    is_property = False
    if v.owners():
        is_property = True

    return dict(
        esl_info=dict(
            is_property=is_property,
            type_ref=type_inst_map[v.variable.the_type].name.tok_text,
            comments=v.comments,
        )
    )


def make_component_node(
    c: ComponentInstance, inst_name: str, params: Dict[ElementaryVarNode, InstNode]
) -> Node:
    """Node creation for a component.

    Arguments:
        c: The component for which a node is created.
        name: The instantiation name of the node.
        params: List of parameters of the component.

    Returns:
        Node of kind "component"
    """
    a = _make_component_annotations(c=c, params=params)
    return Node(name=inst_name, kind="component", annotations=a)


def _make_component_annotations(
    c: ComponentInstance, params: Dict[ElementaryVarNode, InstNode]
) -> Dict[str, Dict[str, Any]]:
    """Component annotation creation.

    Arguments:
        c: The component for which the annotations must be created.
        params: List of parameters of the component.

    Returns:
        See YAML output spec /components/schemas/ComponentAnnotation
    """
    property_variables = [
        v.name
        for v in params.values()
        if v.owners() and set(v.owners()).intersection(c.compdef.parameters)
    ]
    return dict(
        esl_info=dict(
            definition_name=c.def_name_tok.tok_text,
            property_variables=property_variables,
            comments=c.comments,
        )
    )


def make_need_node(n: Need, inst_name: str) -> Node:
    """Node creation for a component.

    Arguments:
      n: The need for which a node must be created.
      name: The instation name of the need.
      inst_prefix: Instantiation prefix

    Returns:
      Node of kind "need"
    """
    name = f"{inst_name}.{n.label_tok.tok_text}"
    a = _make_need_annotations(n=n, inst_prefix=inst_name)
    return Node(name=name, kind="need", annotations=a)


def _make_need_annotations(n: Need, inst_prefix: str) -> Dict[str, Dict[str, Any]]:
    """Need annotation creation.

    Arguments:
      n: The need for which the annotations dict must be created.
      inst_prefix: The instantiation prefix.

    Returns:
      See YAML output spec /components/schemas/ComponentAnnotation.
    """
    return dict(
        esl_info=dict(
            subject=inst_prefix + "." + n.subject_tok.tok_text,
            text=n.description,
            comments=n.comments,
        )
    )


def make_goal_node(g: Goal, inst_name: str, inst_map: Dict[ElementaryVarNode, InstNode]) -> Node:
    """Goal node creation.

    Arguments:
      g: The goal for which a node must be created.
      inst_name: The instantiation name of the component.
      inst_map: Dictionary containing the instantiated variables.

    Returns:
      Node of kind "function".
    """
    name = inst_name + "." + g.label_tok.tok_text
    a = _make_goal_annotations(g=g, inst_name=inst_name, inst_map=inst_map)
    return Node(name=name, kind="function_spec", annotations=a)


def _make_goal_annotations(
    g: Goal, inst_name: str, inst_map: Dict[ElementaryVarNode, InstNode]
) -> Dict[str, Dict[str, Any]]:
    """Goal annotation creation.

    Arguments:
      g: The goal for which a node must be created.
      inst_name: The instantiation name of the component.
      inst_map: Dictionary containing the instantiated variables.

    Returns:
      See YAML output spec /components/schemas/FunctionSpecAnnotation.
    """
    body = dict(
        active=inst_name + "." + g.active.tok_text,
        auxiliary=g.doesaux.tok_text,
        verb=g.verb.tok_text,
        variables=_get_variable_inst_names(inst_map, g.flows),
        preposition=g.prepos.tok_text,
        passive=inst_name + "." + g.passive.tok_text,
        subclauses=_make_subclause_annotations(g, inst_map),
    )

    return dict(esl_info=dict(sub_kind="goal", form=g.goal_kind, body=body, comments=g.comments))


def make_transform_node(
    t: Transformation, inst_name: str, inst_map: Dict[ElementaryVarNode, InstNode]
) -> Node:
    """Transformation node creation.

    Arguments:
      t: The transformation for which a node must be created.
      inst_name: The instantiation name of the component.
      inst_map: Dictionary containing the instantiated variables.

    Returns:
      Node of kind "function_spec".
    """
    name = inst_name + "." + t.label_tok.tok_text
    a = _make_transform_annotations(t=t, inst_name=inst_name, inst_map=inst_map)
    return Node(name=name, kind="function_spec", annotations=a)


def _unfold_compound_var(v: CompoundVarNode) -> List[ElementaryVarNode]:
    """Unfolding CompoundVarNode into a list of ElementartyVarNode"""
    vrs = []
    for v in v.child_nodes:
        if isinstance(v, ElementaryVarNode):
            vrs.append(v)
        elif isinstance(v, CompoundVarNode):
            vrs += _unfold_compound_var(v)

    return vrs


def _get_variable_inst_names(
    inst_map: Dict[ElementaryVarNode, InstNode], flows: List[Flow]
) -> List[str]:
    """Converting list of Flow objects into a list of instantiated variable names."""
    vrs = []
    for f in flows:
        if isinstance(f.flow_node, ElementaryVarNode):
            vrs.append(inst_map.get(f.flow_node).name)
        elif isinstance(f.flow_node, CompoundVarNode):
            vrs += [inst_map[v].name for v in _unfold_compound_var(f.flow_node)]

    return vrs


def _make_transform_annotations(
    t: Transformation, inst_name: str, inst_map: Dict[ElementaryVarNode, InstNode]
) -> Dict[str, Dict[str, Any]]:
    """Transform annotation creation.

    Arguments:
      t: The transformation for which a node must be created.
      inst_name: The instantiation name of the component.
      inst_map: Dictionary containing the instantiated variables.

    Returns:
      See YAML output spec /components/schemas/FunctionSpecAnnotation.
    """
    body = dict(
        active=inst_name,
        auxiliary=t.doesaux_tok.tok_text,
        verb=t.verb_tok.tok_text,
        input_variables=_get_variable_inst_names(inst_map, t.in_flows),
        preposition=t.prepos_tok.tok_text,
        output_variables=_get_variable_inst_names(inst_map, t.out_flows),
        subclauses=_make_subclause_annotations(t, inst_map),
    )

    return dict(
        esl_info=dict(
            sub_kind="transformation",
            form=t.transform_kind,
            body=body,
            comments=t.comments,
        )
    )


def make_design_node(
    d: Design, inst_name: str, inst_map: Dict[ElementaryVarNode, InstNode]
) -> Node:
    """Design spec node creation.

    Arguments:
      d: The design specification for which a node must be created.
      inst_name: The instantiation name of the component.
      inst_map: Dictionary containing the instantiated variables.

    Returns:
      Node of kind "design_spec".
    """
    n = inst_name + "." + d.label_tok.tok_text
    a = _make_design_annotations(d=d, inst_map=inst_map)
    return Node(name=n, kind="design_spec", annotations=a)


def _make_design_annotations(
    d: Design, inst_map: Dict[ElementaryVarNode, InstNode]
) -> Dict[str, Dict[str, any]]:
    """Design spec annotation creation.

    Arguments:
      d: The design spec for which a node must be created.
      inst_map: Dictionary containing the instantiated variables.

    Returns:
      See YAML output spec /components/schemas/DesignSpecAnotation.
    """
    body = _make_design_rule_annotation(expr=d.expr, inst_map=inst_map)
    subs = _make_subclause_annotations(d, inst_map)
    return dict(esl_info=dict(form=d.design_kind, body=body, sub_clauses=subs, comments=d.comments))


def _make_subclause_annotations(
    e: Union[Goal, Transformation, Design], inst_map: Dict[ElementaryVarNode, InstNode]
) -> List[Dict[str, Any]]:
    """Subclause spec annotation creation.

    Arguments:
      e: The relation for which annotations must be created.
      inst_map: Dictionary containing the instantiated variables.

    Returns:
      List of YAML output spec /components/schemas/SubclauseSpecAnnotation.
    """
    a = []
    for s in e.sub_clauses:
        name = s.label_tok.tok_text
        form = "requirement"

        # if s.expr.is_constraint:
        #    form = "constraint"

        body = _make_design_rule_annotation(s.expr, inst_map)

        a.append(dict(name=name, form=form, body=body))

    return a


comparison_dict = {
    "least": "at least",
    "most": "at most",
    "equal": "equal to",
    "not": "not equal to",
    "smaller": "smaller than",
    "greater": "greater than",
    "approximately": "approximately",
}

aux_dict = {
    "must": "must be",
    "should": "should be",
    "could": "could be",
    "would": "would be",
    "won't": "won't be",
    "shall": "shall be",
    "is": "is",
}


def _make_design_rule_annotation(
    expr: Union[exprs.RelationComparison, exprs.Disjunction, exprs.ObjectiveComparison],
    inst_map: Dict[ElementaryVarNode, InstNode],
) -> List[Dict[str, Any]]:
    """Design rule spec annotation creation.

    Arguments:
      expr: The the expression for which the annotations must be created.
      inst_map: Dictionary containing the instantiated variables.

    Returns:
      See YAML output spec /components/schemas/DesignRuleSpecArray.
    """
    if isinstance(expr, exprs.RelationComparison):
        return [
            dict(
                subject=inst_map[expr.lhs_var.var_node].name,
                auxiliary=aux_dict[expr.isaux_tok.tok_text],
                comparison=comparison_dict[expr.cmp_tok.tok_text],
                bound=_make_bound_annotation(expr.rhs_varval, inst_map),
            )
        ]
    elif isinstance(expr, exprs.ObjectiveComparison):
        if expr.maximize:
            comparison = "maximized"
        else:
            comparison = "minimized"

        return [
            dict(
                subject=inst_map[expr.lhs_var.var_node].name,
                auxiliary=expr.aux_tok.tok_text,
                comparison=comparison,
                bound=None,
            )
        ]
    elif isinstance(expr, exprs.Disjunction):
        a = []
        for c in expr.childs:
            a.extend(_make_design_rule_annotation(expr=c, inst_map=inst_map))
        return a
    else:
        return []


def make_behavior_node(
    b: BehaviorFunction, inst_name: str, inst_map: Dict[ElementaryVarNode, InstNode]
) -> Node:
    """Behavior spec node creation.

    Arguments:
      d: The behavior specification for which a node must be created.
      inst_name: The instantiation name of the component.
      inst_map: Dictionary containing the instantiated variables.

    Returns:
      Node of kind "behavior_spec".
    """
    n = inst_name + "." + b.name_tok.tok_text
    a = _make_behavior_annotations(b=b, inst_map=inst_map)
    return Node(name=n, kind="behavior_spec", annotations=a)


def _make_behavior_annotations(
    b: BehaviorFunction, inst_map: Dict[ElementaryVarNode, InstNode]
) -> Dict[str, Dict[str, Any]]:
    """Behavior spec annotation creation.

    Arguments:
      h: The behavior spec for which the annotations must be created.
      inst_map: Dictionary containing the instantiated variables.

    Returns:
      See YAML output spec /components/schemas/BehaviorSpecAnnotation.
    """
    default = []
    if b.default_results:
        default = [
            dict(
                name=r.name_tok.tok_text,
                body=_make_design_rule_annotation(r.result, inst_map),
            )
            for r in b.default_results
        ]

    return dict(
        esl_info=dict(
            form=b.behavior_kind,
            cases=[_make_case_annotations(c, inst_map) for c in b.cases],
            default=default,
            comments=b.comments,
        )
    )


def _make_case_annotations(
    c: BehaviorFunction, inst_map: Dict[ElementaryVarNode, InstNode]
) -> Dict[str, Any]:
    """Case spec annotation creation.

    Arguments:
      r: The behavior specication for which the case annotations must be created.
      inst_map: Dictionary containing the instantiated variables.

    Returns:
      See YAML output spec /components/schemas/CaseSpec.
    """
    return dict(
        name=c.name_tok.tok_text,
        when_clauses=[
            dict(
                name=r.name_tok.tok_text,
                body=_make_design_rule_annotation(r.comparison, inst_map),
            )
            for r in c.conditions
        ],
        then_clauses=[
            dict(
                name=r.name_tok.tok_text,
                body=_make_design_rule_annotation(r.result, inst_map),
            )
            for r in c.results
        ],
    )


def make_relation_node(
    r: RelationInstance, inst_name: str, inst_map: Dict[ElementaryVarNode, InstNode]
) -> Node:
    """Relation spec node creation.

    Arguments:
      d: The relation specification for which a node must be created.
      inst_name: The instantiation name of the component.
      inst_map: Dictionary containing the instantiated variables.

    Returns:
      Node of kind "relation_spec".
    """
    n = inst_name + "." + r.inst_name_tok.tok_text
    a = _make_relation_annotations(r, inst_map)
    return Node(name=n, kind="relation_spec", annotations=a)


def _get_arg_inst_names(
    inst_map: Dict[ElementaryVarNode, InstNode], args: List[InstanceArgument]
) -> List[str]:
    """Converting list of InstanceArgument objects into a list of instantiated variable names."""
    vrs = []
    for a in args:
        if isinstance(a.argnode, ElementaryVarNode):
            vrs.append(inst_map.get(a.argnode).name)
        elif isinstance(a.argnode, CompoundVarNode):
            vrs += [inst_map[v].name for v in _unfold_compound_var(a.argnode)]

    return vrs


def _make_relation_annotations(
    r: RelationInstance, inst_map: Dict[ElementaryVarNode, InstNode]
) -> Dict[str, Dict[str, Any]]:
    """Relation spec annotation creation.

    Arguments:
      r: The relation for which a node must be created.
      inst_map: Dictionary containing the instantiated variables.

    Returns:
      See YAML output spec /components/schemas/RelationSpecAnnotation.
    """
    reqv = []
    retv = []
    relv = []
    for a, p in zip(r.arguments, r.reldef.params):
        if p.direction == "input":
            reqv.extend(_get_arg_inst_names(inst_map, a))
        elif p.direction == "output":
            retv.extend(_get_arg_inst_names(inst_map, a))
        elif p.direction == "inp_out":
            relv.extend(_get_arg_inst_names(inst_map, a))

    return dict(
        esl_info=dict(
            definition_name=r.def_name_tok.tok_text,
            required_variables=reqv,
            returned_variables=retv,
            related_variables=relv,
            comments=r.comments,
        )
    )


def _make_bound_annotation(
    varval: Union[exprs.Value, exprs.VariableValue],
    inst_map: Dict[ElementaryVarNode, InstNode],
) -> Dict[str, str]:
    """Bound annotation creation

    Arguments:
      varval: The value or variable that denotes the bound.
      inst_map:  Dictionary containing the instantiated variables.

    Returns:
      See YAML output spec /components/schemas/Bound.
    """
    if isinstance(varval, exprs.Value):
        if varval.unit:
            return dict(value=varval.value.tok_text, unit=varval.unit.tok_text)
        else:
            return dict(value=varval.value.tok_text, unit=None)

    assert isinstance(varval, exprs.VariableValue)

    return dict(value=inst_map[varval.var_node].name, unit=None)


TAG_COMMENT_PAT = re.compile(r"^\s*@(?P<tag>[\S]+)[ \t]+(?P<comment>.*)")


def post_process_comments(nodes: List[Node]) -> None:
    """Post-processing comments attached to nodes.

    Arguments:
        nodes: List of nodes for which the comments must be post-processed.

    Note:
        This is a simple implementation to process documentation tags as described
        in LEP0008.
    """
    for n in nodes:
        cms = n.annotations.esl_info.get("comments", [])
        plain_comments = []
        tagged_comments = defaultdict(list)

        for cm in cms:
            match = TAG_COMMENT_PAT.match(cm)
            if match is None:
                plain_comments.append(cm)
            else:
                groups = match.groupdict()
                tagged_comments[groups["tag"]].append(groups["comment"])

        n.annotations.esl_info["comments"] = plain_comments
        n.annotations.esl_info["tagged_comments"] = tagged_comments



"""Line matcher state machines for relation instances."""
from typing import TYPE_CHECKING

from raesl.compile.machine_files import argument_list, typing

if TYPE_CHECKING:
    from raesl.compile.typechecking.ast_builder import AstBuilder


_RELATION_HEADER_SPEC = """
relation_header:
    start initial;
    start -> s1 [RELATION_KW] tag=relation;

    end accept=relation_header;
    s1 -> end [NL_TK];
    s1 -> end [EOF_TK];
"""


def _process_relinst_header(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    builder.notify_new_section(tags["relation"][0], False)


_RELATION_INSTANCE_SPEC = """
relation_instance:
    start initial;
    start -> s1 [NAME] tag=relation_inst_name;
    s1 -> s2 [COLON_TK];
    s2 -> s3 [NAME] tag=relation_def_name;

    end accept=instance_line;
    s3 -> end [NL_TK];
    s3 -> end [EOF_TK];
"""


def _process_new_relinst(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    current_comp = builder.compdef_builder.current_component
    assert current_comp is not None
    current_comp.new_relinst(tags["relation_inst_name"][0], tags["relation_def_name"][0])


_RELATION_ARGUMENT_HEADER_SPEC = """
relation_argument_header:
    start initial;
    start -> s1 [REQUIRING_KW] tag=argtype;
    start -> s1 [RETURNING_KW] tag=argtype;
    start -> s1 [RELATING_KW] tag=argtype;

    s1 -> s2 [ARGUMENT_KW];

    end accept=header_argument_line;
    s2 -> end [NL_TK];
    s2 -> end [EOF_TK];
"""


def _process_relinst_argheader(
    tags: typing.TokensDict, _accept: str, builder: "AstBuilder"
) -> None:
    current_comp = builder.compdef_builder.current_component
    assert current_comp is not None
    current_comp.relinst_argheader(tags["argtype"][0])


# Argument line processing.
def _process_instance_arguments(
    tags: typing.TokensDict, _accept: str, builder: "AstBuilder"
) -> None:
    arguments = argument_list.process_argument_list_line(tags)
    current_comp = builder.compdef_builder.current_component
    assert current_comp is not None
    current_comp.add_relinst_arguments(arguments)


MACHINES: typing.MachineTripletList = [
    ("RELATION_HEADER_MACHINE", _RELATION_HEADER_SPEC, _process_relinst_header),
    ("RELATION_INSTANCE_MACHINE", _RELATION_INSTANCE_SPEC, _process_new_relinst),
    (
        "RELATION_ARGUMENT_HEADER_MACHINE",
        _RELATION_ARGUMENT_HEADER_SPEC,
        _process_relinst_argheader,
    ),
    (
        "RELATION_ARGUMENT_MACHINE",
        argument_list.ARGUMENT_LINE_SPEC,
        _process_instance_arguments,
    ),
]



"""Edge building for the instantiated output graph.

Reference:
https://ratio-case.gitlab.io/docs/reference/esl_reference/dependency-derivations.html
"""
from collections import defaultdict
from copy import deepcopy
from itertools import chain
from typing import Any, Dict, Generator, Iterable, List, Optional, Set, Tuple

from raesl import logger
from raesl.compile import diagnostics
from raesl.compile.instantiating.node_building import NodeStore
from ragraph.edge import Edge
from ragraph.node import Node


class EdgeStore:
    """Edge storage with multiple catagories for quicker access to specific subsets."""

    categories = ["edges"]  # Updated in EdgeFactory init.

    def __init__(self):
        for cat in self.categories:
            setattr(self, cat, list())

    def clear(self):
        """Clear all edge categories."""
        for cat in self.categories:
            getattr(self, cat).clear()

    def add(self, edge: Edge, *args):
        """Add node to :obj:`self.nodes` and any other specified maps in args."""
        self.edges.append(edge)
        for m in args:
            getattr(self, m).append(edge)
        logger.debug(f"Added ragraph.edge.Edge {edge} to edges and {args}.")

    def consume(self, edges: Iterable[Edge], *args):
        """Add any edges from an iterable to given categories."""
        for e in edges:
            self.add(e, *args)


class EdgeFactory:
    def __init__(
        self,
        diag_store: diagnostics.DiagnosticStore,
        node_store: Optional[NodeStore] = None,
    ):
        self.diag_store = diag_store
        self.node_store = node_store

        # Set categories dynamically.
        EdgeStore.categories = ["edges"] + [i[6:] for i in dir(self) if i.startswith("_make_")]
        self.edge_store = EdgeStore()

    def _add(self, edge: Edge, *args):
        """Proxy for :obj:`EdgeStore.add`."""
        self.edge_store.add(edge, *args)

    def _consume(self, edges: Iterable[Edge], *args):
        """Proxy for :obj:`EdgeStore.consume`."""
        self.edge_store.consume(edges, *args)

    def _make(self, cat: str) -> List[Edge]:
        """Make an edge category if it is still empty and return it for quick access."""
        if not getattr(self.edge_store, cat):
            func = getattr(self, f"_make_{cat}")
            func()

        return getattr(self.edge_store, cat)

    def make_edges(self, node_store: Optional[NodeStore] = None) -> List[Edge]:
        """Derive edges from a :obj:`NodeStore` object."""
        # Clear edge lists
        self.edge_store.clear()

        # Set provided node_store or try an keep the current one.
        self.node_store = self.node_store if node_store is None else node_store
        if self.node_store is None:
            return list()

        for cat in self.edge_store.categories:
            # Make edges if list is empty. This guarantees each category is only
            # generated once regardless of them calling eachother internally.
            if cat == "edges":
                continue
            self._make(cat)

        return self.edge_store.edges

    def _make_evf(self):
        """Compute functional dependencies between variable nodes."""
        nodes = self.node_store.nodes
        tfs = [
            n
            for n in self.node_store.transforms.values()
            if not nodes[n.annotations.esl_info.get("body").get("active")].children
        ]  # Only transforms where the active component doesn't have children.

        for t in tfs:
            for vi in t.annotations.esl_info["body"]["input_variables"]:
                for vj in t.annotations.esl_info["body"]["output_variables"]:
                    self._add(
                        Edge(
                            source=nodes[vi],
                            target=nodes[vj],
                            kind="functional_dependency",
                            labels=[nodes[vi].annotations.esl_info["type_ref"]],
                            annotations=dict(
                                esl_info=dict(reason=dict(function_specifications=[t.name]))
                            ),
                        ),
                        "evf",
                    )

    def _make_evb(self):
        """Compute logical dependencies between variables for behavior specs."""
        for b in self.node_store.behaviors.values():
            for c in b.annotations.esl_info["cases"]:
                self._consume(
                    _generate_evbc(behavior=b, case=c, nodes=self.node_store.nodes),
                    "evb",
                )

    def _make_evd(self):
        """Compute design dependencies between variables."""
        nodes = self.node_store.nodes
        designs = self.node_store.designs.values()
        relations = self.node_store.relations.values()

        for d in designs:
            for r in d.annotations.esl_info["body"]:
                vi = nodes[r["subject"]]

                bound = r.get("bound")
                if bound is None:
                    # Mim / max requirement
                    continue

                vj = nodes.get(bound["value"])

                if not vj:
                    continue

                self._add(
                    Edge(
                        source=vi,
                        target=vj,
                        kind="design_dependency",
                        labels=[vi.annotations.esl_info["type_ref"]],
                        annotations=dict(
                            esl_info=dict(reason=dict(design_specifications=[d.name]))
                        ),
                    ),
                    "evd",
                )

                self._add(
                    Edge(
                        source=vj,
                        target=vi,
                        kind="design_dependency",
                        labels=[vj.annotations.esl_info["type_ref"]],
                        annotations=dict(
                            esl_info=dict(reason=dict(design_specifications=[d.name]))
                        ),
                    ),
                    "evd",
                )

        for r in relations:
            for vi in (
                r.annotations.esl_info["required_variables"]
                + r.annotations.esl_info["related_variables"]
            ):
                for vj in (
                    r.annotations.esl_info["returned_variables"]
                    + r.annotations.esl_info["related_variables"]
                ):
                    if vi == vj:
                        continue

                    self._add(
                        Edge(
                            source=nodes[vi],
                            target=nodes[vj],
                            kind="design_dependency",
                            labels=[nodes[vi].annotations.esl_info["type_ref"]],
                            annotations=dict(
                                esl_info=dict(reason=dict(relation_specifications=[r.name]))
                            ),
                        ),
                        "evd",
                    )

    def _make_ecf(self):
        """Compute function dependencies between component nodes."""
        nodes = self.node_store.nodes
        transforms = self.node_store.transforms.values()
        goals = self.node_store.goals.values()

        nt_dict = defaultdict(list)
        for t in transforms:
            nt_dict[t.annotations.esl_info["body"]["active"]].append(t)

        for g in goals:
            src = nodes[g.annotations.esl_info["body"]["active"]]
            trg = nodes[g.annotations.esl_info["body"]["passive"]]

            lbs = [
                nodes[v].annotations.esl_info["type_ref"]
                for v in g.annotations.esl_info["body"]["variables"]
            ]

            self._add(
                Edge(
                    source=src,
                    target=trg,
                    kind="functional_dependency",
                    labels=lbs,
                    annotations=dict(esl_info=dict(reason=dict(function_specifications=[g.name]))),
                ),
                "ecf",
            )

            self._add(
                Edge(
                    source=trg,
                    target=src,
                    kind="functional_dependency",
                    labels=lbs,
                    annotations=dict(esl_info=dict(reason=dict(function_specifications=[g.name]))),
                ),
                "ecf",
            )

            gvars = set(g.annotations.esl_info["body"]["variables"])
            for c in src.descendants:
                for t in nt_dict[c.name]:
                    t_out_vars = set(t.annotations.esl_info["body"]["output_variables"])

                    shared_v = gvars.intersection(t_out_vars)
                    if not shared_v:
                        continue

                    lbs = [nodes[v].annotations.esl_info["type_ref"] for v in shared_v]
                    self._add(
                        Edge(
                            source=c,
                            target=trg,
                            kind="functional_dependency",
                            labels=lbs,
                            annotations=dict(
                                esl_info=dict(reason=dict(function_specifications=[g.name, t.name]))
                            ),
                        ),
                        "ecf",
                    )

                    self._add(
                        Edge(
                            source=trg,
                            target=c,
                            kind="functional_dependency",
                            labels=lbs,
                            annotations=dict(
                                esl_info=dict(reason=dict(function_specifications=[g.name, t.name]))
                            ),
                        ),
                        "ecf",
                    )

            for c in trg.descendants:
                for t in nt_dict[c.name]:
                    t_in_vars = set(t.annotations.esl_info["body"]["input_variables"])

                    shared_v = gvars.intersection(t_in_vars)
                    if not shared_v:
                        continue

                    lbs = [nodes[v].annotations.esl_info["type_ref"] for v in shared_v]
                    self._add(
                        Edge(
                            source=src,
                            target=c,
                            kind="functional_dependency",
                            labels=lbs,
                            annotations=dict(
                                esl_info=dict(reason=dict(function_specifications=[t.name, g.name]))
                            ),
                        ),
                        "ecf",
                    )

                    self._add(
                        Edge(
                            source=c,
                            target=src,
                            kind="functional_dependency",
                            labels=lbs,
                            annotations=dict(
                                esl_info=dict(reason=dict(function_specifications=[t.name, g.name]))
                            ),
                        ),
                        "ecf",
                    )

        for ti in transforms:
            for tj in transforms:
                ci = ti.annotations.esl_info["body"]["active"]
                cj = tj.annotations.esl_info["body"]["active"]

                if ti is tj or ci == cj:
                    continue

                ti_in_vars = set(ti.annotations.esl_info["body"]["input_variables"])
                ti_out_vars = set(ti.annotations.esl_info["body"]["output_variables"])
                tj_in_vars = set(tj.annotations.esl_info["body"]["input_variables"])
                tj_out_vars = set(tj.annotations.esl_info["body"]["output_variables"])
                shared_vars_ij = ti_out_vars.intersection(tj_in_vars)
                shared_vars_ji = tj_out_vars.intersection(ti_in_vars)

                if shared_vars_ij:
                    lbs = [nodes[v].annotations.esl_info["type_ref"] for v in shared_vars_ij]
                    self._add(
                        Edge(
                            source=nodes[ci],
                            target=nodes[cj],
                            kind="functional_dependency",
                            labels=lbs,
                            annotations=dict(
                                esl_info=dict(
                                    reason=dict(function_specifications=[ti.name, tj.name])
                                )
                            ),
                        ),
                        "ecf",
                    )

                    self._add(
                        Edge(
                            source=nodes[cj],
                            target=nodes[ci],
                            kind="functional_dependency",
                            labels=lbs,
                            annotations=dict(
                                esl_info=dict(
                                    reason=dict(function_specifications=[ti.name, tj.name])
                                )
                            ),
                        ),
                        "ecf",
                    )

                if shared_vars_ji:
                    lbs = [nodes[v].annotations.esl_info["type_ref"] for v in shared_vars_ji]
                    self._add(
                        Edge(
                            source=nodes[cj],
                            target=nodes[ci],
                            kind="functional_dependency",
                            labels=lbs,
                            annotations=dict(
                                esl_info=dict(
                                    reason=dict(function_specifications=[tj.name, ti.name])
                                )
                            ),
                        ),
                        "ecf",
                    )

                    self._add(
                        Edge(
                            source=nodes[ci],
                            target=nodes[cj],
                            kind="functional_dependency",
                            labels=lbs,
                            annotations=dict(
                                esl_info=dict(
                                    reason=dict(function_specifications=[tj.name, ti.name])
                                )
                            ),
                        ),
                        "ecf",
                    )

    def _make_ecb(self):
        """Compute logical dependencies between components."""
        nodes = self.node_store.nodes
        goals = self.node_store.goals.values()
        transforms = self.node_store.transforms.values()

        evb = self._make("evb")
        ev_dict = defaultdict(list)
        for e in evb:
            ev_dict[(e.source, e.target)].append(e)

        mfv = self._make("mfv")
        vf_dict = defaultdict(list)
        for e in mfv:
            vf_dict[e.target.name].append(e.source)

        # Check goal-goal behavior dependencies.
        for sources in [goals, transforms]:
            for targets in [goals, transforms]:
                for gi in sources:
                    for gj in targets:
                        edge = _get_component_behavior_dependency(
                            nodes=nodes,
                            src=gi,
                            trg=gj,
                            ev_dict=ev_dict,
                            vf_dict=vf_dict,
                        )
                        if edge is not None:
                            self._add(edge, "ecb")

    def _make_ecd(self):
        """Create design dependencies between components."""
        nodes = self.node_store.nodes
        cmps = self.node_store.components.values()

        evd = self._make("evd")

        e_dict = {(e.source.name, e.target.name): e for e in evd}
        for ci in cmps:
            for pi in ci.annotations.esl_info["property_variables"]:
                for cj in cmps:
                    for pj in cj.annotations.esl_info["property_variables"]:
                        if (pi, pj) not in e_dict:
                            continue

                        # Migrate edge up tree
                        for ai in [ci] + ci.ancestors:
                            for aj in [cj] + cj.ancestors:
                                if ai == aj:
                                    continue
                                self._add(
                                    Edge(
                                        source=ai,
                                        target=aj,
                                        kind="design_dependency",
                                        labels=[nodes[pi].annotations.esl_info["type_ref"]],
                                        annotations=dict(
                                            esl_info=dict(
                                                reason=e_dict[(pi, pj)].annotations.esl_info[
                                                    "reason"
                                                ]
                                            )
                                        ),
                                    ),
                                    "ecd",
                                )

    def _make_eft(self):
        """Create traceability dependencies between function specifications."""
        nodes = self.node_store.nodes
        cps = self.node_store.components.values()
        gls = self.node_store.goals.values()
        tfs = self.node_store.transforms.values()

        ctd: Dict[str, List[Node]] = defaultdict(list)
        for t in tfs:
            ctd[t.annotations.esl_info["body"]["active"]].append(t)

        cgd: Dict[str, List[Node]] = defaultdict(list)
        for g in gls:
            cgd[g.annotations.esl_info["body"]["active"]].append(g)

        for c in cps:
            if not c.children or not ctd.get(c.name):
                continue

            child_ts: List[Node] = []
            child_gs: List[Node] = []
            for child in c.children:
                child_ts.extend(ctd[child.name])
                child_gs.extend(cgd[child.name])

            evf_c = []

            for t in child_ts:
                for vi in t.annotations.esl_info["body"]["input_variables"]:
                    for vj in t.annotations.esl_info["body"]["output_variables"]:
                        evf_c.append(
                            Edge(
                                source=nodes[vi],
                                target=nodes[vj],
                                kind="functional_dependency",
                                labels=[nodes[vi].annotations.esl_info["type_ref"]],
                                annotations=dict(
                                    esl_info=dict(reason=dict(function_specifications=[t.name]))
                                ),
                            )
                        )

            for t in ctd[c.name]:
                srcs = [nodes[v] for v in t.annotations.esl_info["body"]["input_variables"]]
                trgs = [nodes[v] for v in t.annotations.esl_info["body"]["output_variables"]]

                paths = _get_paths_between_all(srcs=srcs, trgs=set(trgs), edges=evf_c)

                pathvars = set()
                for path in paths:
                    pathvars.update(set(path))

                for child_t in child_ts:
                    shared_path_variables = set(
                        child_t.annotations.esl_info["body"]["output_variables"]
                    ).intersection(pathvars)
                    if shared_path_variables:
                        self._add(
                            Edge(
                                source=t,
                                target=child_t,
                                kind="traceability_dependency",
                                annotations=dict(
                                    esl_info=dict(
                                        reason=dict(path_variables=list(shared_path_variables))
                                    )
                                ),
                            ),
                            "eft",
                        )

                for child_g in child_gs:
                    shared_path_variables = set(
                        child_g.annotations.esl_info["body"]["variables"]
                    ).intersection(pathvars)
                    if shared_path_variables:
                        self._add(
                            Edge(
                                source=t,
                                target=child_g,
                                kind="traceability_dependency",
                                annotations=dict(
                                    esl_info=dict(
                                        reason=dict(path_variables=list(shared_path_variables))
                                    )
                                ),
                            ),
                            "eft",
                        )

    def _make_eff(self):
        """Creating all functional dependencies between between goals and transforms."""
        nodes = self.node_store.nodes
        gls = self.node_store.goals.values()
        tfs = self.node_store.transforms.values()

        for t1 in tfs:
            for t2 in tfs:
                if (
                    t1.annotations.esl_info["body"]["active"]
                    != t2.annotations.esl_info["body"]["active"]
                ):
                    continue

                vout = set(t1.annotations.esl_info["body"]["output_variables"])
                vin = set(t2.annotations.esl_info["body"]["input_variables"])
                v_shared = vout.intersection(vin)

                if not v_shared:
                    continue

                lbs = [nodes[v].annotations.esl_info["type_ref"] for v in v_shared]
                self._add(
                    Edge(
                        source=t1,
                        target=t2,
                        kind="functional_dependency",
                        labels=lbs,
                        annotations=dict(
                            esl_info=dict(reason=dict(shared_variables=list(v_shared)))
                        ),
                    ),
                    "eff",
                )

            for g in gls:
                tvin = set(t1.annotations.esl_info["body"]["input_variables"])
                tvout = set(t1.annotations.esl_info["body"]["output_variables"])
                gv = set(g.annotations.esl_info["body"]["variables"])

                vin_shared = tvin.intersection(gv)
                if vin_shared:
                    lbs = [nodes[v].annotations.esl_info["type_ref"] for v in vin_shared]
                    self._add(
                        Edge(
                            source=g,
                            target=t1,
                            kind="functional_dependency",
                            labels=lbs,
                            annotations=dict(
                                esl_info=dict(reason=dict(shared_variables=list(vin_shared)))
                            ),
                        ),
                        "eff",
                    )
                vout_shared = tvout.intersection(gv)
                if vout_shared:
                    lbs = [nodes[v].annotations.esl_info["type_ref"] for v in vout_shared]
                    self._add(
                        Edge(
                            source=t1,
                            target=g,
                            kind="functional_dependency",
                            labels=lbs,
                            annotations=dict(
                                esl_info=dict(reason=dict(shared_variables=list(vout_shared)))
                            ),
                        ),
                        "eff",
                    )

    def _make_efb(self):
        """Compute behavior dependencies between goal- and transformation specs."""
        nodes = self.node_store.nodes
        gls = self.node_store.goals.values()
        tfs = self.node_store.transforms.values()

        evb = self._make("evb")

        evdict = defaultdict(list)
        for e in evb:
            evdict[(e.source, e.target)].append(e)

        mfv = self._make("mfv")
        vf_dict = defaultdict(list)
        for e in mfv:
            vf_dict[e.target.name].append(e.source)

        for t1 in tfs:
            for t2 in tfs:
                if (
                    t1.annotations.esl_info["body"]["active"]
                    != t2.annotations.esl_info["body"]["active"]
                ):
                    continue

                edge = _get_function_behavior_dependency(
                    nodes=nodes, src=t1, trg=t2, ev_dict=evdict, vf_dict=vf_dict
                )
                if edge is not None:
                    self._add(edge, "efb")

            for g in gls:
                edge = _get_function_behavior_dependency(
                    nodes=nodes, src=t1, trg=g, ev_dict=evdict, vf_dict=vf_dict
                )
                if edge is not None:
                    self._add(edge, "efb")

                edge = _get_function_behavior_dependency(
                    nodes=nodes, src=g, trg=t1, ev_dict=evdict, vf_dict=vf_dict
                )
                if edge is not None:
                    self._add(edge, "efb")

    def _make_ehb(self):
        """Creating logical dependencies between behavior specifications."""
        nodes = self.node_store.nodes
        bhs = self.node_store.behaviors.values()

        for bi in bhs:
            for bj in bhs:
                if bi == bj:
                    continue

                vis = []
                for case in bi.annotations.esl_info["cases"]:
                    for tc in case["then_clauses"]:
                        vis.extend([nodes[r["subject"]] for r in tc["body"]])

                vjs = []
                for case in bj.annotations.esl_info["cases"]:
                    for wc in case["when_clauses"]:
                        vjs.extend([nodes[r["subject"]] for r in wc["body"]])

                shared_vars = set(vis).intersection(set(vjs))
                if not shared_vars:
                    continue

                lbs = [v.annotations.esl_info["type_ref"] for v in shared_vars]
                self._add(
                    Edge(
                        source=bi,
                        target=bj,
                        kind="logical_dependency",
                        labels=lbs,
                        annotations=dict(
                            esl_info=dict(
                                reason=dict(shared_variables=[v.name for v in shared_vars])
                            )
                        ),
                    ),
                    "ehb",
                )

    def _make_ehc(self):
        """Creating coordination dependencies between behavior specifications."""
        nodes = self.node_store.nodes
        bhs = self.node_store.behaviors.values()

        for bi in bhs:
            for bj in bhs:
                if bi == bj:
                    continue

                vis = []
                for case in bi.annotations.esl_info["cases"]:
                    for tc in case["then_clauses"]:
                        vis.extend([nodes[r["subject"]] for r in tc["body"]])

                vjs = []
                for case in bj.annotations.esl_info["cases"]:
                    for tc in case["then_clauses"]:
                        vjs.extend([nodes[r["subject"]] for r in tc["body"]])

                shared_vars = set(vis).intersection(set(vjs))
                if not shared_vars:
                    continue

                lbs = [v.annotations.esl_info["type_ref"] for v in shared_vars]
                self._add(
                    Edge(
                        source=bi,
                        target=bj,
                        kind="coordination_dependency",
                        labels=lbs,
                        annotations=dict(
                            esl_info=dict(
                                reason=dict(shared_variables=[v.name for v in shared_vars])
                            )
                        ),
                    ),
                    "ehc",
                )

    def _make_edc(self):
        """Create coordination dependencies between design specifications."""
        nodes = self.node_store.nodes
        dss = self.node_store.designs.values()

        for di in dss:
            vis = set()
            for r in di.annotations.esl_info["body"]:
                vis.add(nodes[r["subject"]])
            for dj in dss:
                if di == dj:
                    continue
                vjs = set()
                for r in dj.annotations.esl_info["body"]:
                    vjs.add(nodes[r["subject"]])

                shared_vars = vis.intersection(vjs)
                if not shared_vars:
                    continue

                self._add(
                    Edge(
                        source=di,
                        target=dj,
                        kind="coordination_dependency",
                        labels=[v.annotations.esl_info["type_ref"] for v in shared_vars],
                        annotations=dict(
                            esl_info=dict(
                                reason=(dict(shared_variables=[v.name for v in shared_vars]))
                            )
                        ),
                    ),
                    "edc",
                )

    def _make_enc(self):
        """Create coordination dependencies between needs."""
        nds = self.node_store.needs.values()
        for ni in nds:
            for nj in nds:
                if ni == nj:
                    continue

                if not ni.annotations.esl_info["subject"] == nj.annotations.esl_info["subject"]:
                    continue

                self._add(
                    Edge(
                        source=ni,
                        target=nj,
                        kind="mapping_dependency",
                        labels=["shared subject"],
                        annotations=dict(
                            esl_info=dict(
                                reason=dict(shared_subject=ni.annotations.esl_info["subject"])
                            )
                        ),
                    ),
                    "enc",
                )

    def _make_erc(self):
        """Create coordination dependencies between relations."""
        nodes = self.node_store.nodes
        rls = self.node_store.relations.values()

        for ri in rls:
            for rj in rls:
                if ri == rj:
                    continue

                vris = set(
                    ri.annotations.esl_info["related_variables"]
                    + ri.annotations.esl_info["returned_variables"]
                )

                vrjs = set(
                    rj.annotations.esl_info["required_variables"]
                    + rj.annotations.esl_info["related_variables"]
                )

                shared_vars = vris.intersection(vrjs)
                if not shared_vars:
                    continue

                self._add(
                    Edge(
                        source=ri,
                        target=rj,
                        kind="coordination_dependency",
                        labels=[nodes[v].annotations.esl_info["type_ref"] for v in shared_vars],
                        annotations=dict(esl_info=dict(reason=dict(shared_vars=list(shared_vars)))),
                    ),
                    "erc",
                )

    def _make_mcf(self):
        """Compute mapping relations between components and functions."""
        nodes = self.node_store.nodes
        n2c = self.node_store.components
        gls = self.node_store.goals.values()
        tfs = self.node_store.transforms.values()
        ctd = defaultdict(list)

        for t in tfs:
            ctd[nodes[t.annotations.esl_info["body"]["active"]]].append(t)
            lbs = [
                nodes[v].annotations.esl_info["type_ref"]
                for v in t.annotations.esl_info["body"]["input_variables"]
                + t.annotations.esl_info["body"]["output_variables"]
            ]

            self._add(
                Edge(
                    source=n2c[t.annotations.esl_info["body"]["active"]],
                    target=t,
                    labels=lbs,
                    kind="mapping_dependency",
                ),
                "mcf",
            )

        for g in gls:
            gvs = set([nodes[v] for v in g.annotations.esl_info["body"]["variables"]])
            lbs = [v.annotations.esl_info["type_ref"] for v in gvs]
            a = n2c[g.annotations.esl_info["body"]["active"]]
            p = n2c[g.annotations.esl_info["body"]["passive"]]

            self._add(Edge(source=a, target=g, labels=lbs, kind="mapping_dependency"), "mcf")
            self._add(Edge(source=p, target=g, labels=lbs, kind="mapping_dependency"), "mcf")

            for d in a.descendants:
                if not ctd.get(d):
                    continue

                for t in ctd[d]:
                    outvars = set(
                        nodes[v] for v in t.annotations.esl_info["body"]["output_variables"]
                    )
                    shared_vars = outvars.intersection(gvs)

                    if shared_vars:
                        self._add(
                            Edge(
                                source=d,
                                target=g,
                                labels=[v.annotations.esl_info["type_ref"] for v in shared_vars],
                                kind="mapping_dependency",
                            ),
                            "mcf",
                        )

            for d in p.descendants:
                for t in ctd[d]:
                    invars = set(
                        nodes[v] for v in t.annotations.esl_info["body"]["input_variables"]
                    )
                    shared_vars = invars.intersection(gvs)

                    if shared_vars:
                        self._add(
                            Edge(
                                source=d,
                                target=g,
                                labels=[v.annotations.esl_info["type_ref"] for v in shared_vars],
                                kind="mapping_dependency",
                            ),
                            "mcf",
                        )

    def _make_mcv(self):
        """Create mapping relations between components and variables."""
        nodes = self.node_store.nodes
        cmps = self.node_store.components.values()
        mcf = self._make("mcf")

        # Get mapping relations from function specs.
        for m in mcf:
            f = m.target
            if m.target.annotations.esl_info["sub_kind"] == "goal":
                vrs = m.target.annotations.esl_info["body"]["variables"]
            if m.target.annotations.esl_info["sub_kind"] == "transformation":
                vrs = (
                    f.annotations.esl_info["body"]["input_variables"]
                    + f.annotations.esl_info["body"]["output_variables"]
                )

            for v in vrs:
                self._add(
                    Edge(
                        source=m.source,
                        target=nodes[v],
                        labels=[nodes[v].annotations.esl_info["type_ref"]],
                        annotations=dict(
                            esl_info=dict(reason=dict(function_specifications=[f.name]))
                        ),
                        kind="mapping_dependency",
                    ),
                    "mcv",
                )

        # Get mapping relations from properties
        for c in cmps:
            for v in c.annotations.esl_info["property_variables"]:
                self._add(
                    Edge(
                        source=c,
                        target=nodes[v],
                        labels=[nodes[v].annotations.esl_info["type_ref"]],
                        annotations=dict(esl_info=dict(reason={})),
                        kind="mapping_dependency",
                    ),
                    "mcv",
                )

    def _make_mcb(self):
        """Create mapping dependencies between components and behavior specs."""
        nodes = self.node_store.nodes
        tss = self.node_store.transforms.values()
        cmps = self.node_store.components.values()
        bhs = self.node_store.behaviors.values()

        cvd = defaultdict(set)
        for t in tss:
            c = nodes[t.annotations.esl_info["body"]["active"]]
            vs = t.annotations.esl_info["body"]["output_variables"]
            for v in vs:
                cvd[c, v].add(v)

        for c in cmps:
            for v in c.annotations.esl_info["property_variables"]:
                cvd[c, v].add(v)

        bvd = defaultdict(set)
        for b in bhs:
            for case in b.annotations.esl_info["cases"]:
                for tc in case["then_clauses"]:
                    for v in [r["subject"] for r in tc["body"]]:
                        bvd[b, v].add(v)

        shared_var_dict = defaultdict(list)
        for c, v in cvd:
            for b in bhs:
                if not bvd.get((b, v)):
                    continue
                shared_var_dict[c, b].append(v)

        for (c, b), vs in shared_var_dict.items():
            self._add(
                Edge(
                    source=c,
                    target=b,
                    kind="mapping_dependency",
                    labels=[nodes[v].annotations.esl_info["type_ref"] for v in vs],
                ),
                "mcb",
            )

    def _make_mfb(self):
        """Create mapping dependencies between function specs and behavior specs."""
        bhs = self.node_store.behaviors.values()
        mfv = self._make("mfv")
        mbv = self._make("mbv")

        mfvd = {}
        for e in mfv:
            mfvd[e.source, e.target] = e
        mbvd = {}
        for e in mbv:
            mbvd[e.source, e.target] = e

        shared_var_dict = defaultdict(list)

        for f, v in mfvd:
            for b in bhs:
                if not mbvd.get((b, v)):
                    continue
                shared_var_dict[f, b].append(v)

        for (f, b), vs in shared_var_dict.items():
            self._add(
                Edge(
                    source=f,
                    target=b,
                    kind="mapping_dependency",
                    labels=[v.annotations.esl_info["type_ref"] for v in vs],
                ),
                "mfb",
            )

    def _make_mbv(self):
        """Create mapping relations between behavior specifications and variables."""
        nodes = self.node_store.nodes
        bhs = self.node_store.behaviors.values()
        for b in bhs:
            vs = []
            for case in b.annotations.esl_info["cases"]:
                for wc in case["when_clauses"]:
                    vs.extend([nodes[r["subject"]] for r in wc["body"]])

                for tc in case["then_clauses"]:
                    vs.extend([nodes[r["subject"]] for r in tc["body"]])

            for v in vs:
                self._add(
                    Edge(
                        source=b,
                        target=v,
                        kind="mapping_dependency",
                        labels=[v.annotations.esl_info["type_ref"]],
                    ),
                    "mbv",
                )

    def _make_mcd(self):
        """Create mapping relations between components and design requirements."""
        nodes = self.node_store.nodes
        cmps = self.node_store.components.values()
        dss = self.node_store.designs.values()
        mcv = self._make("mcv")

        cvd: Dict[str, Set[str]] = defaultdict(set)
        for e in mcv:
            cvd[e.source].add(e.target)

        for c in cmps:
            for d in dss:
                vds = set()
                for r in d.annotations.esl_info["body"]:
                    vds.add(nodes[r["subject"]])

                    bound = r.get("bound")
                    if bound is None:
                        # Mim / max requirement
                        continue

                    if nodes.get(bound["value"]):
                        vds.add(nodes[bound["value"]])

                    shared_vars = cvd[c].intersection(vds)
                    if not shared_vars:
                        continue

                    self._add(
                        Edge(
                            source=c,
                            target=d,
                            kind="mapping_dependency",
                            labels=[v.annotations.esl_info["type_ref"] for v in shared_vars],
                            annotations=dict(
                                esl_info=dict(
                                    reason=dict(shared_variables=[v.name for v in shared_vars])
                                )
                            ),
                        ),
                        "mcd",
                    )

    def _make_mcr(self):
        """Create mapping relations between components and relations."""
        nodes = self.node_store.nodes
        rls = self.node_store.relations.values()
        mcv = self._make("mcv")

        cvd = defaultdict(set)
        for e in mcv:
            cvd[e.source].add(e.target.name)

        rvd = defaultdict(set)
        for r in rls:
            rvd[r] = set(
                r.annotations.esl_info["required_variables"]
                + r.annotations.esl_info["related_variables"]
                + r.annotations.esl_info["returned_variables"]
            )

        for c, cvs in cvd.items():
            for r, rvs in rvd.items():
                shared_vars = cvs.intersection(rvs)
                if not shared_vars:
                    continue

                self._add(
                    Edge(
                        source=c,
                        target=r,
                        kind="mapping_dependency",
                        labels=[nodes[v].annotations.esl_info["type_ref"] for v in shared_vars],
                        annotations=dict(esl_info=dict(reason=dict(shared_vars=list(shared_vars)))),
                    ),
                    "mcr",
                )

    def _make_mfv(self):
        """Construct mapping dependencies between function specs and variables."""
        nodes = self.node_store.nodes
        tfs = self.node_store.transforms.values()
        gls = self.node_store.goals.values()

        for t in tfs:
            vs = [
                nodes[v]
                for v in t.annotations.esl_info["body"]["input_variables"]
                + t.annotations.esl_info["body"]["output_variables"]
            ]
            for v in vs:
                self._add(
                    Edge(
                        source=t,
                        target=v,
                        kind="mapping_dependency",
                        labels=[v.annotations.esl_info["type_ref"]],
                    ),
                    "mfv",
                )

        for g in gls:
            vs = [nodes[v] for v in g.annotations.esl_info["body"]["variables"]]
            for v in vs:
                self._add(
                    Edge(
                        source=g,
                        target=v,
                        kind="mapping_dependency",
                        labels=[v.annotations.esl_info["type_ref"]],
                    ),
                    "mfv",
                )

    def _make_mfd(self):
        """Construct mapping dependencies between function specs and design specs."""
        nodes = self.node_store.nodes
        gls = self.node_store.goals.values()
        tfs = self.node_store.transforms.values()
        dss = self.node_store.designs.values()

        fvd: Dict[str, Set[str]] = defaultdict(set)
        for g in gls:
            fvd[g] |= set(g.annotations.esl_info["body"]["variables"])
        for t in tfs:
            fvd[t] |= set(
                t.annotations.esl_info["body"]["input_variables"]
                + t.annotations.esl_info["body"]["output_variables"]
            )

        dvd: Dict[str, Set[str]] = defaultdict(set)
        for d in dss:
            for r in d.annotations.esl_info["body"]:
                dvd[d].add(r["subject"])

        for f in chain(gls, tfs):
            for d in dss:
                shared_vars = fvd[f].intersection(dvd[d])
                if not shared_vars:
                    continue

                self._add(
                    Edge(
                        source=f,
                        target=d,
                        kind="mapping_dependency",
                        labels=[nodes[v].annotations.esl_info["type_ref"] for v in shared_vars],
                        annotations=dict(esl_info=dict(reason=dict(shared_vars=list(shared_vars)))),
                    ),
                    "mfd",
                )

    def _make_mfr(self):
        """Create mapping dependencies between functions specs and relation specs."""
        nodes = self.node_store.nodes
        gls = self.node_store.goals.values()
        tfs = self.node_store.transforms.values()
        rls = self.node_store.relations.values()

        fvd: Dict[str, Set[str]] = defaultdict(set)
        for g in gls:
            fvd[g] |= set(g.annotations.esl_info["body"]["variables"])

        for t in tfs:
            fvd[t] |= set(
                t.annotations.esl_info["body"]["input_variables"]
                + t.annotations.esl_info["body"]["output_variables"]
            )

        rvd: Dict[str, Set[str]] = defaultdict(set)
        for r in rls:
            rvd[r] = set(
                r.annotations.esl_info["required_variables"]
                + r.annotations.esl_info["related_variables"]
                + r.annotations.esl_info["returned_variables"]
            )

        for f, fvs in fvd.items():
            for r, rvs in rvd.items():
                shared_vars = fvs.intersection(rvs)
                if not shared_vars:
                    continue

                self._add(
                    Edge(
                        source=f,
                        target=r,
                        kind="mapping_dependency",
                        labels=[nodes[v].annotations.esl_info["type_ref"] for v in shared_vars],
                        annotations=dict(esl_info=dict(reason=dict(shared_vars=list(shared_vars)))),
                    ),
                    "mfr",
                )

    def _make_mbd(self):
        """Create mapping dependencies between behavior requirements and design
        specifications.
        """
        nodes = self.node_store.nodes
        bhs = self.node_store.behaviors.values()
        dss = self.node_store.designs.values()

        bvd: Dict[Node, Set[str]] = defaultdict(set)
        for b in bhs:
            for case in b.annotations.esl_info["cases"]:
                for tc in case["then_clauses"]:
                    bvd[b] |= set([r["subject"] for r in tc["body"]])

        dvd: Dict[Node, Set[str]] = defaultdict(set)
        for d in dss:
            for r in d.annotations.esl_info["body"]:
                dvd[d].add(r["subject"])

        for b, bvs in bvd.items():
            for d, dvs in dvd.items():
                shared_vars = bvs.intersection(dvs)
                if shared_vars:
                    self._add(
                        Edge(
                            source=b,
                            target=d,
                            kind="mapping_dependency",
                            labels=[nodes[v].annotations.esl_info["type_ref"] for v in shared_vars],
                            annotations=dict(
                                esl_info=dict(reason=dict(shared_vars=list(shared_vars)))
                            ),
                        ),
                        "mbd",
                    )

    def _make_mbr(self):
        """Create mapping dependencies between behavior requirements and relation
        specifications.
        """
        nodes = self.node_store.nodes
        bhs = self.node_store.behaviors.values()
        rls = self.node_store.relations.values()

        bvd: Dict[Node, Set[str]] = defaultdict(set)
        for b in bhs:
            for case in b.annotations.esl_info["cases"]:
                for tc in case["then_clauses"]:
                    bvd[b] |= set([r["subject"] for r in tc["body"]])
                for wc in case["when_clauses"]:
                    bvd[b] |= set([r["subject"] for r in wc["body"]])

        rvd: Dict[Node, Set[str]] = defaultdict(set)
        for r in rls:
            rvd[r] = set(
                r.annotations.esl_info["required_variables"]
                + r.annotations.esl_info["related_variables"]
                + r.annotations.esl_info["returned_variables"]
            )

        for b, bvs in bvd.items():
            for r, rvs in rvd.items():
                shared_vars = bvs.intersection(rvs)
                if shared_vars:
                    self._add(
                        Edge(
                            source=b,
                            target=r,
                            kind="mapping_dependency",
                            labels=[nodes[v].annotations.esl_info["type_ref"] for v in shared_vars],
                            annotations=dict(
                                esl_info=dict(reason=dict(shared_vars=list(shared_vars)))
                            ),
                        ),
                        "mbr",
                    )

    def _make_mvd(self):
        """Create mapping dependencies between variables and design specifications."""
        nodes = self.node_store.nodes
        dss = self.node_store.designs.values()

        for d in dss:
            dvs = []
            for r in d.annotations.esl_info["body"]:
                dvs.append(r["subject"])

                bound = r.get("bound")
                if bound is None:
                    # Mim / max requirement
                    continue

                if nodes.get(bound["value"]):
                    dvs.append(bound["value"])

            for v in dvs:
                self._add(
                    Edge(
                        source=nodes[v],
                        target=d,
                        kind="mapping_dependency",
                        labels=[nodes[v].annotations.esl_info["type_ref"]],
                    ),
                    "mvd",
                )

    def _make_mvr(self):
        """Create mapping relations between variables and relation specifications."""
        nodes = self.node_store.nodes
        rls = self.node_store.relations.values()

        for r in rls:
            rvs = (
                r.annotations.esl_info["required_variables"]
                + r.annotations.esl_info["related_variables"]
                + r.annotations.esl_info["returned_variables"]
            )

            for v in rvs:
                self._add(
                    Edge(
                        source=nodes[v],
                        target=r,
                        kind="mapping_dependency",
                        labels=[nodes[v].annotations.esl_info["type_ref"]],
                    ),
                    "mvr",
                )

    def _make_mdr(self):
        """Create mapping dependencies between design specs and relation specs."""
        nodes = self.node_store.nodes
        dss = self.node_store.designs.values()
        rls = self.node_store.relations.values()

        dvd: Dict[Node, Set[str]] = defaultdict(set)
        for d in dss:
            for r in d.annotations.esl_info["body"]:
                dvd[d].add(r["subject"])

                bound = r.get("bound")
                if bound is None:
                    # Mim / max requirement
                    continue

                if nodes.get(bound["value"]):
                    dvd[d].add(bound["value"])

        rvd: Dict[Node, Set[str]] = defaultdict(set)
        for r in rls:
            rvd[r] = set(
                r.annotations.esl_info["required_variables"]
                + r.annotations.esl_info["related_variables"]
                + r.annotations.esl_info["returned_variables"]
            )

        for d, dvs in dvd.items():
            for r, rvs in rvd.items():
                shared_vars = dvs.intersection(rvs)
                if shared_vars:
                    self._add(
                        Edge(
                            source=d,
                            target=r,
                            kind="mapping_dependency",
                            labels=[nodes[v].annotations.esl_info["type_ref"] for v in shared_vars],
                            annotations=dict(
                                esl_info=dict(reason=dict(shared_vars=list(shared_vars)))
                            ),
                        ),
                        "mdr",
                    )

    def _make_mnx(self):
        """Create mapping dependencies between needs and all other elements."""
        nodes = self.node_store.nodes
        nds = self.node_store.needs.values()
        for n in nds:
            s = nodes[n.annotations.esl_info["subject"]]
            self._add(
                Edge(source=s, target=n, kind="mapping_dependency", labels=["mapping"]),
                "mnx",
            )


def _generate_evbc(
    behavior: Node, case: Dict[str, Any], nodes: Dict[str, Node]
) -> Generator[Edge, None, None]:
    """Compute logical dependencies between variables for a behavior case specification.

    Arguments:
        behavior: Behavior specification that contains the case from which the logical
            dependencies must be derived.
        case: The case from which the logical dependencies must be derived.
        nodes: Dictionary of node name to Node object.

    Yields:
        Corresponding edges.
    """
    wv = []
    for wc in case["when_clauses"]:
        wv.extend([r["subject"] for r in wc["body"]])

    tv = []
    for tc in case["then_clauses"]:
        tv.extend([r["subject"] for r in tc["body"]])

    for vi in wv:
        for vj in tv:
            yield Edge(
                source=nodes[vi],
                target=nodes[vj],
                kind="logical_dependency",
                labels=[nodes[vi].annotations.esl_info["type_ref"]],
                annotations=dict(
                    esl_info=dict(reason=dict(behavior_specifications=[behavior.name]))
                ),
            )


def _get_component_behavior_dependency(
    nodes: Dict[str, Node],
    src: Node,
    trg: Node,
    ev_dict: Dict[Tuple[Node, Node], List[Edge]],
    vf_dict: Dict[str, List[Node]],
) -> Optional[Edge]:
    """Check if two component nodes have a behavior dependency.

    Arguments:
        nodes: Dictionary from node to node name.
        src: The source node for which the dependency must be checked.
        trg: THe target node for which the dependency must be checked.
        ev_dict: Dictionary of node pairs to list of edges.
        vf_dict: Dictionary of variable name to function spec nodes.

    Returns:
        An Edge if a behavior dependency exists. Otherwise None.
    """
    vis, vjs = [], []
    if src.annotations.esl_info["sub_kind"] == "goal":
        vis = [nodes[v] for v in src.annotations.esl_info["body"]["variables"]]
    elif src.annotations.esl_info["sub_kind"] == "transformation":
        vis = [nodes[v] for v in src.annotations.esl_info["body"]["output_variables"]]
    if trg.annotations.esl_info["sub_kind"] == "goal":
        vjs = [nodes[v] for v in trg.annotations.esl_info["body"]["variables"]]
    elif trg.annotations.esl_info["sub_kind"] == "transformation":
        vjs = [nodes[v] for v in trg.annotations.esl_info["body"]["output_variables"]]

    # Add properties
    ci = nodes[src.annotations.esl_info["body"]["active"]]
    cj = nodes[trg.annotations.esl_info["body"]["active"]]

    if ci == cj:
        return None

    if ci.annotations.esl_info.get("property_variables"):
        vis += [nodes[v] for v in ci.annotations.esl_info["property_variables"]]

    if cj.annotations.esl_info.get("property_variables"):
        vjs += [nodes[v] for v in cj.annotations.esl_info["property_variables"]]

    reasons = set()
    vrs = []
    edges = []
    for v in ev_dict.values():
        edges.extend(v)

    paths = _get_paths_between_all(srcs=vis, trgs=vjs, edges=edges)
    for path in paths:
        if len(path) == 2:
            vi, vj = path[0], path[1]
            edges = ev_dict.get((nodes[vi], nodes[vj]), [])
            for e in edges:
                reasons.update(e.annotations.esl_info["reason"]["behavior_specifications"])
            vrs.extend([nodes[path[0]], nodes[path[-1]]])
        else:
            # Check id none of the intermediate path variables relate to a function
            # spec. If so proceed.
            transformation_check = True
            for v in path[1:-1]:
                if v in vf_dict:
                    transformation_check = False
                    break

            if transformation_check:
                for vi, vj in zip(path[:-1], path[1:]):
                    edges = ev_dict.get((nodes[vi], nodes[vj]))
                    for e in edges:
                        reasons.update(e.annotations.esl_info["reason"]["behavior_specifications"])

            vrs.extend([nodes[path[0]], nodes[path[-1]]])

    if reasons:
        lbs = [v.annotations.esl_info["type_ref"] for v in vrs]
        return Edge(
            source=ci,
            target=cj,
            kind="logical_dependency",
            labels=lbs,
            annotations=dict(
                esl_info=dict(
                    reason=dict(
                        behavior_specifications=sorted(reasons),
                        function_specifications=[src.name, trg.name],
                    )
                )
            ),
        )
    else:
        return None


def _get_paths_between_all(srcs=List[Node], trgs=List[Node], edges=List[Edge]) -> List[List[str]]:
    """Compute paths between nodes. Based on depth first search.

    Arguments:
         srcs: List of starting nodes of paths.
         trgs: Set of ending nodes of path.
         edges: List of edges between nodes.

    Yields:
        List of lists of node names.
    """
    paths = []

    ed: Dict[str, Dict[str, Edge]] = defaultdict(dict)
    for e in edges:
        ed[e.source][e.target] = e

    for src in srcs:
        paths.extend(_get_paths(src=src, edge_dct=ed, trgs=trgs, visited=[]))

    return paths


def _get_paths(
    src: Node,
    edge_dct: Dict[Node, Dict[Node, Edge]],
    trgs: Set[Node],
    visited: List[Node] = [],
) -> List[List[str]]:
    """Collection all paths (list of node names) between the source node and the set of
    target nodes.

    Arguments:
        src: The source nodes where all paths should start.
        edge_dct: Dictionary of Node to Node to Edge. Contains the edges to be
            considered when searching for paths.
        trgs: Set of node where the paths should end.
        visited: List of nodes already visited. Required to prevent running in cycles.

    Returns:
        List of lists of node names.
    """
    visited.append(src.name)
    paths: List[List[str]] = []
    edges = edge_dct.get(src)
    if not edges:
        return paths

    for n in edges:
        if n.name in visited:
            # node has already been visited. Loop entered
            continue
        elif n in trgs:
            # A target has been reached
            visited.append(n.name)
            paths.append(visited)
        else:
            new_path = deepcopy(visited)
            paths.extend(_get_paths(src=n, edge_dct=edge_dct, trgs=trgs, visited=new_path))

    return paths


def _get_function_behavior_dependency(
    nodes: Dict[str, Node],
    src: Node,
    trg: Node,
    ev_dict: Dict[Tuple[Node, Node], Edge],
    vf_dict: Dict[str, List[Node]],
) -> Optional[Edge]:
    """Check if two function spec nodes have a behavior dependency.

    Arguments:
        nodes: Dictionary from node to node name.
        src: The source node for which the dependency must be checked.
        trg: THe target node for which the dependency must be checked.
        ev_dict: Dictionary of node pairs to edge.
        vf_dict: Dictionary of variable names to function specs.

    Returns:
        An Edge if a behavior dependency exists. Otherwise None.
    """
    vis, vjs = [], []
    if src.annotations.esl_info["sub_kind"] == "goal":
        vis = [nodes[v] for v in src.annotations.esl_info["body"]["variables"]]
    elif src.annotations.esl_info["sub_kind"] == "transformation":
        vis = [nodes[v] for v in src.annotations.esl_info["body"]["output_variables"]]
    if trg.annotations.esl_info["sub_kind"] == "goal":
        vjs = [nodes[v] for v in trg.annotations.esl_info["body"]["variables"]]
    elif trg.annotations.esl_info["sub_kind"] == "transformation":
        vjs = [nodes[v] for v in trg.annotations.esl_info["body"]["output_variables"]]

    reasons = set()
    vrs = []
    edges = []
    for v in ev_dict.values():
        edges.extend(v)

    paths = _get_paths_between_all(srcs=vis, trgs=vjs, edges=edges)
    for path in paths:
        if len(path) == 2:
            vi, vj = path[0], path[1]
            edges = ev_dict.get((nodes[vi], nodes[vj]), [])
            for e in edges:
                reasons.update(e.annotations.esl_info["reason"]["behavior_specifications"])
            vrs.extend([nodes[path[0]], nodes[path[-1]]])
        else:
            # Check id none of the intermediate path variables relate to a function
            # spec. If so proceed.
            transformation_check = True
            for v in path[1:-1]:
                if v in vf_dict:
                    transformation_check = False
                    break

            if transformation_check:
                for vi, vj in zip(path[:-1], path[1:]):
                    edges = ev_dict.get((nodes[vi], nodes[vj]))
                    for e in edges:
                        reasons.update(e.annotations.esl_info["reason"]["behavior_specifications"])

            vrs.extend([nodes[path[0]], nodes[path[-1]]])

    if reasons:
        lbs = [v.annotations.esl_info["type_ref"] for v in vrs]
        return Edge(
            source=src,
            target=trg,
            kind="logical_dependency",
            labels=lbs,
            annotations=dict(esl_info=dict(reason=dict(behavior_specifications=sorted(reasons)))),
        )
    else:
        return None



"""Code for handling documentation comments in AST objects.

The code has two main concepts, namely DocStore which expresses the capability
to find elements that can store doc comments, and DocAddElement which expresses
the capability to actually store such comments.

In general thus, storing a doc comment thus involves two steps. First the correct
element is searched based on a supplied (possibly dotted) name, then the
comment is actually stored if an element could be found.

As the above is pretty minimal, several additional classes exist.
- The DocElement can both store comments and also provide them again.
  (Compound element can't store comments, but forward them to their children and
  thus have no way to retrieve the comments given to them.)
- The DefaultDocElement implements DocElement.
- The DefaultDocStore implements a DocStore as well as DocElement for non-dotted
  names, providing a simple way to add doc comment storage to many of the
  language elements.

Internally, some more classes exist.

- The ProxyDocStore that acts as DocStore elsewhere in the specification. Its
  primary use is in comment sections, pointing to the actual elements to receive
  the doc comments of the section.

- The DummyElement acts as a DocStore for any doc comment, and reports a warning
  about them being ignored. These elements act as barrier and guard against doc
  comments being added to a previous element where that should not happen.
  For example, doc comments after a 'define type' line, should not be added
  to the definitions above that line.

- Finally, DocCommentDistributor implements the distribution of doc comments to all
  elements of the specification after type-checking has been performed.
"""
import itertools
from collections import defaultdict
from typing import Generator, List, Optional

from raesl.compile import diagnostics
from raesl.compile.scanner import Token


class DocAddElement:
    """Interface class of an element that can store doc comments."""

    def add_comment(self, comment_tok: Token) -> None:
        """Add found documentation comment.

        Arguments:
            comment_tok: The raw documentation token to add.
        """
        raise NotImplementedError("Implement me in {}".format(repr(self)))


class DocElement(DocAddElement):
    """Interface class of an element that can store and retrieve doc comments."""

    def get_comment(self) -> List[str]:
        """Retrieve the stored comments."""
        raise NotImplementedError("Implement me in {}".format(repr(self)))


class DefaultDocElement(DocElement):
    """Default implementation for storing and retrieving doc comments.

    Arguments:
        comments: The comments themselves, non-empty text after dropping the leading
            '#<' and surrounding white-space.
    """

    def __init__(self):
        super(DefaultDocElement, self).__init__()
        self.comments: List[str] = []

    def add_comment(self, comment_tok: Token):
        """Add found documentation comment.

        Arguments:
            comment_tok: The raw documentation token to add.
        """
        self.comments.extend(decode_doc_comments(comment_tok))

    def get_comment(self) -> List[str]:
        return self.comments


def decode_doc_comments(comment_tok: Token) -> List[str]:
    """Convert a doc comment token to the containing description text.

    Arguments:
        comment_tok: Token with the documentation comment.

    Returns:
        The text (for as far as it exists).
    """
    assert comment_tok.tok_text.startswith("#<")
    text = comment_tok.tok_text[2:].strip()
    if text:
        return [text]
    return []


class DocStore:
    """Interface class that can find where to store doc comments for a given name.
    If doc_tok is None, the element does not get any documentation comments.

    Arguments:
        doc_tok: Token defining the position of the element in the input for
            documenting. Documentation comments after this position and before any
            other existing DocStore.doc_tok will be attached to this element.
    """

    def __init__(self, doc_tok: Optional[Token]):
        self.doc_tok = doc_tok

    def resolve_element(self, name: str) -> Optional[DocAddElement]:
        """Try to find the documentation element indicated by its (dotted) name.

        Arguments:
            name: Name of the element to find.

        Returns:
            The documentation element associated with the provided name if it can be
                resolved.
        """
        raise NotImplementedError("Implement me in {}".format(repr(self)))

    def get_error_position(self, name: str) -> int:
        """Return the index in the given string where an error occurs in resolving the
        name.

        Arguments:
            name: Name of the element to find.

        Returns:
            Approximated index in the string where matching the element fails.
                Returned value has no meaning if resolving succeeds.
        """
        return 0


class DefaultDocStore(DocStore, DefaultDocElement):
    """Class that can store and retrieve doc-comments for non-dotted names.

    Arguments:
        doc_tok: Token defining the position of the element in the input for
            documenting. Documentation comments after this position and before any
            other existing DocStore.doc_tok will be attached to this element.
    """

    def __init__(self, doc_tok: Optional[Token]):
        # Can't use super() due to different argument lists.
        DocStore.__init__(self, doc_tok)
        DefaultDocElement.__init__(self)

    def resolve_element(self, name: str):
        if self.doc_tok is None:
            return None
        if "." in name:
            return None
        if self.doc_tok.tok_text != name:
            return None

        return self

    def get_error_position(self, name: str):
        i = name.find(".")
        if i >= 0:
            return i
        return 0


class ProxyDocStore(DocStore):
    """Proxy element that represents a real DocStore element except at a different
    position in the specification. This is useful in 'comment' sections where names of
    elements are provided that exist elsewhere in the component.

    Arguments:
        doc_tok: Token defining the position of the element in the input for
            documenting. Documentation comments after this position and before any
            other existing DocStore.doc_tok will be attached to this element.
        real_element: Real element which is symbolically at the 'doc_tok' position, too.
    """

    def __init__(self, doc_tok: Token, real_element: DocStore):
        super(ProxyDocStore, self).__init__(doc_tok)
        self.real_element = real_element

    def resolve_element(self, name: str):
        return self.real_element.resolve_element(name)

    def get_error_position(self, name: str):
        return self.real_element.get_error_position(name)


class DummyElement(DocStore, DocElement):
    """Class for catching documentation elements that have no proper owner. Used for
    reporting warnings that such documentation is ignored.

    Arguments:
        doc_tok: Token defining the position of the element in the input for
            documenting. Documentation comments after this position and before any
            other existing DocStore.doc_tok will be attached to this element.
        report_errors: Whether to add errors to the problem storage.
    """

    def __init__(self, doc_tok: Token, report_errors: bool = True):
        super(DummyElement, self).__init__(doc_tok)
        self.raw_comments: List[Token] = []
        self.report_errors = report_errors

    def resolve_element(self, name: str):
        # We don't actually care about the name, as anything ending up here is reported
        # as error.
        return self

    def get_error_position(self, name: str):
        # resolve_element() never fails, so the returned value has no meaning anyway.
        return 0

    def add_comment(self, comment_tok: Token):
        # Do not convert to text to preserve the precise position of the comment.
        self.raw_comments.append(comment_tok)

    def get_comment(self):
        assert False, "Comments are not usable, call self.report() instead."

    def report(self, diag_store: diagnostics.DiagnosticStore):
        """Report a warning about all received documentation comments.

        Arguments:
            diag_store: Storage for reported diagnostics.
        """
        if not self.raw_comments or not self.report_errors:
            return

        locs = [tok.get_location() for tok in self.raw_comments]
        diag_store.add(diagnostics.W300(location=locs[0], comments=locs))


class DocCommentDistributor:
    """Class for assigning documentation comments to relevant elements in the
    specification.

    Arguments:
        diag_store: Storage for reported diagnostics.

    Attributes:
        elements: Elements interested in receiving documentation comments.
        dummy_elements: Elements that catch documentation comments without a proper
            home, for warning the user about such comments.
    """

    def __init__(self, diag_store: diagnostics.DiagnosticStore):
        self.diag_store = diag_store
        self.elements: List[DocStore] = []
        self.dummy_elements: List[DummyElement] = []

        # This 'tok' violates pretty much all assumptions of Token, do not try to use
        # it outside the distributor context.
        tok = Token(
            tok_type="DUMMY_TK",
            tok_text="",
            fname=None,
            offset=-1,
            line_offset=-1,
            line_num=-1,
        )
        self.add_dummy_element(tok)

    def add_dummy_element(self, doc_tok: Token, report_errors: bool = True):
        """Insert a dummy element based on the provided token.

        Arguments:
            doc_tok: Token that defines the position of the dummy element.
            report_errors: Whether to report errors for comments that get attached to
                the dummy element.
        """
        dds = DummyElement(doc_tok, report_errors)
        self.dummy_elements.append(dds)
        self.add_element(dds)

    def add_element(self, element: DocStore):
        """Add the provided element to the elements interested in getting doc comments.

        Arguments:
            element: Element to add.
        """
        self.elements.append(element)

    def resolve(self, doc_comments: List[Token]):
        """Distribute the provided documentation comments to the interested elements,
        and report any documentation comments that get assigned in a DummyElement,
        as those are at the wrong spot in the specification.

        Arguments:
            doc_comments: Documentation comments found in the input specification.
        """
        if not doc_comments:  # No comments, nothing to do.
            return

        # At this point, there is
        # - the provided doc_comments, a mostly sorted list of 'DOC_COMMENT_TK' tokens
        #   containing the documentation comments.
        # - the self.elements list, the interested elements, including all dummy
        #   elements for reporting about doc comments at weird places. The tokens of the
        #   former must be assigned to the latter based on offset information and file.
        #   All doc comments must be assigned to that element with the largest offset
        #   less than the offset of the comment within the same file.

        # Sort the comments on file name and offset.
        dc_dict = defaultdict(list)
        for dc in doc_comments:
            dc_dict[dc.fname].append(dc)

        for val in dc_dict.values():
            val.sort(key=lambda tok: tok.offset)

        # The tricky part is that the element list may have several elements at the
        # same offset, due to blindly inserting dummy elements. This can be resolved by
        # either being more careful about offsets of dummy elements, or by filtering
        # such duplicates afterwards.
        #
        # The code here does the latter. Sort the interested elements, and filter the
        # dummy duplicate elements out of it. Also append a None element to notify
        # about the end of the list.

        # Sort the elements on file name and offset.
        elm_dict = defaultdict(list)
        for elm in self.elements:
            fname = getattr(elm.doc_tok, "fname", "unknown-file")
            elm_dict[fname].append(elm)

        for key, val in elm_dict.items():
            offsets = [v.doc_tok.offset for v in val]
            sequence = [idx for _, idx in sorted(zip(offsets, range(len(val))))]
            elm_dict[key] = [val[idx] for idx in sequence]

        nondup_elm_dict = {}
        for key, val in elm_dict.items():
            nondup_elm_dict[key] = list(_drop_dup_dummies(val))

        # Last element before the position at 'doc_index'.
        cur_elem: Optional[DocStore] = None

        # Next element after cur_elem, or None if not initialized or at the end.
        next_elem: Optional[DocStore] = None

        for cur_file in nondup_elm_dict:
            if cur_file not in dc_dict:
                continue

            dc_list = dc_dict[cur_file]
            doc_index = 0
            cur_elem = None
            next_elem = None

            for elem in itertools.chain(nondup_elm_dict[cur_file], _gen_none()):
                if cur_elem is None and next_elem is None:
                    # At startup, setup next_elem for the next iteration
                    next_elem = elem
                    # Holds as the distributor inserts a dummy element at offset -1
                    # thus it is always a non-empty list.
                    assert next_elem is not None

                    # Paranoia check, first comment should be at or after that element.
                    comment_offset = dc_list[doc_index].offset
                    assert next_elem.doc_tok
                    if comment_offset < next_elem.doc_tok.offset:
                        loc = dc_list[doc_index].get_location(comment_offset)
                        self.diag_store.add(
                            diagnostics.W300(element=None, location=loc, comments=[loc])
                        )

                    continue

                # Regular iteration
                cur_elem = next_elem
                next_elem = elem  # Might be None
                # Find the document element pointed at by 'cur_elem'.
                assert cur_elem is not None
                assert cur_elem.doc_tok is not None
                cur_doc_element = cur_elem.resolve_element(cur_elem.doc_tok.tok_text)
                if cur_doc_element is None:
                    # Element name cannot be resolved, throw an error and ignore it
                    # further.
                    offset = cur_elem.get_error_position(cur_elem.doc_tok.tok_text)
                    loc = cur_elem.doc_tok.get_location(offset)
                    name = cur_elem.doc_tok.tok_text
                    self.diag_store.add(diagnostics.W300(element=name, location=loc))

                # Process the doc comments belonging to the current element.
                while doc_index < len(dc_list):
                    comment_offset = dc_list[doc_index].offset

                    # If next comment is at or after the next element, cur_elem is done.
                    if (
                        next_elem is not None
                        and next_elem.doc_tok is not None
                        and comment_offset >= next_elem.doc_tok.offset
                    ):
                        break

                    if cur_doc_element is not None:
                        # Reported a problem about failing to resolve already, thus
                        # silently skipping is fine.
                        cur_doc_element.add_comment(dc_list[doc_index])

                    doc_index = doc_index + 1

                # If all comments have been processed, ignore the remaining elements.
                if doc_index >= len(dc_list):
                    break

            # All doc comments distributed, check that nothing ended up at a weird spot.
            for elem in self.dummy_elements:
                elem.report(self.diag_store)


def _drop_dup_dummies(gen: List[DocStore]) -> Generator[DocStore, None, None]:
    """Drop dummy elements at the same offset as a non-dummy element."""
    element = None
    for elm in gen:
        if element is None:
            element = elm
            continue

        if elm.doc_tok.offset > element.doc_tok.offset:
            yield element
            element = elm
            continue

        assert elm.doc_tok.offset == element.doc_tok.offset
        if isinstance(element, DummyElement) and elm.doc_tok.fname == element.doc_tok.fname:
            element = elm  # 'element' is not useful, 'elm' cannot be worse!
            continue

    if element is not None:
        yield element


def _gen_none() -> Generator[None, None, None]:
    """Generate a stream consisting of a single 'None' value."""
    yield None



"""Line matcher state machines for variable declarations."""
from typing import TYPE_CHECKING

from raesl.compile.ast.components import VarParam
from raesl.compile.machine_files import typing

if TYPE_CHECKING:
    from raesl.compile.typechecking.ast_builder import AstBuilder


_VARIABLE_HEADER_SPEC = """
variable_header:
    start initial;
    start -> s1 [VARIABLE_KW] tag=variable;

    end accept=variable_header;
    s1 -> end [NL_TK];
    s1 -> end [EOF_TK];
"""


def _process_varheader(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    builder.notify_new_section(tags["variable"][0], False)


_VARIABLE_LINE_SPEC = """
variable_line:
    start initial;
    start -> s1 [NAME] tag=varname;
    s1 -> start [COMMA_TK];

    s1 -> s2 [IS_KW];
    s2 -> s3 [A_KW];
    s3 -> s4 [NAME] tag=typename;

    end accept=variable_line;
    s4 -> end [NL_TK];
    s4 -> end [EOF_TK];
"""


def _process_varline(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    varnames = tags["varname"]
    typename = tags["typename"][0]
    variables = [VarParam(True, vname, typename) for vname in varnames]
    current_comp = builder.compdef_builder.current_component
    assert current_comp is not None
    current_comp.add_variables(variables)


MACHINES: typing.MachineTripletList = [
    ("VARIABLE_HEADER_MACHINE", _VARIABLE_HEADER_SPEC, _process_varheader),
    ("VARIABLE_LINE_MACHINE", _VARIABLE_LINE_SPEC, _process_varline),
]



"""Line matcher state machines for type definitions."""
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, cast

from raesl.compile.ast import exprs
from raesl.compile.machine_files import typing
from raesl.compile.machine_files.utils import get_one, get_optional, make_loc_names

if TYPE_CHECKING:
    from raesl.compile.scanner import Token
    from raesl.compile.typechecking.ast_builder import AstBuilder


def get_unit_specification_part(start_loc: str, end_1locs: List[str], prefix: str = "") -> str:
    """State machine part implementing the 'unit-specification' rule.

    unit-specification ::=
        "with" ( "unit" | "units" ) UNIT-NAME { "," UNIT-NAME }
    """
    assert len(end_1locs) == 1

    locs = {
        "start": start_loc,
        "end": end_1locs[0],
    }
    locs.update(make_loc_names(prefix, "uspec", 2))

    text = """\
        {uspec1}; {uspec2};

        {start} -> {uspec1} [WITH_KW];
        {uspec1} -> {uspec2} [UNIT_KW] tag=has_unit_spec;
        {uspec2} -> {end} [NONCOMMA] tag=unit_name;
        {end} -> {uspec2} [COMMA_TK];
    """
    return text.format(**locs)


def _convert_unit_spec(tags: Dict[str, List["Token"]]) -> Optional[List["Token"]]:
    """Convert a unit specification to a list of unit names, if it exists."""
    if "has_unit_spec" not in tags:
        return None

    return tags["unit_name"]


def get_enumeration_specification_part(
    start_loc: str, end_2locs: List[str], prefix: str = ""
) -> str:
    """State machine part implementing the 'enumeration-specification' rule.

    enumeration-specification ::=
        "is" "an" "enumeration" "of" VALUE [ UNIT ] { "," VALUE [ UNIT ]}
    """
    assert len(end_2locs) == 2

    lenum_locs = make_loc_names(prefix, "lenum", 2)
    locs = {
        "start": start_loc,
        "end": lenum_locs["lenum2"],
    }
    locs.update(lenum_locs)

    text = """
        {lenum1}; {lenum2};

        {start} -> {lenum1} [IS_KW];
        {lenum1} -> {end} [A_KW];
    """
    text = text.format(**locs)
    return text + get_short_enumeration_specification_part(locs["end"], end_2locs, prefix)


def get_short_enumeration_specification_part(
    start_loc: str, end_2locs: List[str], prefix: str = ""
) -> str:
    """State machine part implementing the short 'enumeration-specification' rule.

    short-enumeration-specification ::=
        "enumeration" "of" VALUE [ UNIT ] { "," VALUE [ UNIT ]}
    """
    assert len(end_2locs) == 2

    locs = {
        "start": start_loc,
        "end1": end_2locs[0],
        "end2": end_2locs[1],
    }
    locs.update(make_loc_names(prefix, "enum", 2))

    text = """\
        {enum1}; {enum2};

        {start} -> {enum1} [ENUMERATION_KW] tag=has_enum_spec;
        {enum1} -> {enum2} [OF_KW];
        {enum2} -> {end1} [NONCOMMA] tag=enum_value;
        {end1} -> {end2} [NONCOMMA] tag=enum_unit;

        {end1} -> {enum2} [COMMA_TK];
        {end2} -> {enum2} [COMMA_TK];
    """
    return text.format(**locs)


def _convert_enum_spec(tags: Dict[str, List["Token"]]) -> Optional[List[exprs.Value]]:
    """Convert an enumeration specification to a number of values if it exists in the
    given tags.

    Arguments:
        tags: Tags from the matched line.

    Returns:
        Values of the enumeration.
    """
    if "has_enum_spec" not in tags:
        return None

    offsets = cast(List[Optional[int]], [tok.offset for tok in tags["enum_value"]]) + [None]

    values = []
    unit_tags = tags.get("enum_unit", [])
    for i, val in enumerate(tags["enum_value"]):
        unit = get_optional(unit_tags, val.offset, offsets[i + 1])
        values.append(exprs.Value(val, unit))
    return values


def get_interval_specification_part(start_loc: str, end_4locs: List[str], prefix: str = "") -> str:
    """State machine part implementing the 'interval-specification' rule.

    interval-specification ::=
        "of" interval { "or" interval }
    """
    assert len(end_4locs) == 4

    locs = {
        "start": start_loc,
        "end1": end_4locs[0],
        "end2": end_4locs[1],
        "end3": end_4locs[2],
        "end4": end_4locs[3],
    }
    locs.update(make_loc_names(prefix, "intval", 6))

    text = """\
        {intval1}; {intval2}; {intval3}; {intval4}; {intval5}; {intval6};

        {start} -> {intval1} [OF_KW] tag=has_intval_spec;

        {intval1} -> {intval2} [AT_KW];
        {intval2} -> {intval3} [LEAST_KW];
        {intval3} -> {end1}    [NONCOMMA] tag=lowerbound_value;
        {end1}    -> {end2}    [NONCOMMA] tag=lowerbound_unit;

        {end1}    -> {intval4} [AND_KW] tag=and;
        {end2}    -> {intval4} [AND_KW] tag=and;

        {intval4} -> {intval5} [AT_KW];
        {intval5} -> {intval6} [MOST_KW];
        {intval6} -> {end3}    [NONCOMMA] tag=upperbound_value;
        {end3}    -> {end4}    [NONCOMMA] tag=upperbound_unit;

        {intval2} -> {intval6} [MOST_KW]; # Skip 'at least'

        {end1}    -> {intval1} [OR_KW] tag=or; # 'or' after lower bound
        {end2}    -> {intval1} [OR_KW] tag=or;

        {end3}   -> {intval1}  [OR_KW] tag=or; # 'or' after upper bound
        {end4}   -> {intval1}  [OR_KW] tag=or;
    """
    return text.format(**locs)


def _convert_interval_spec(
    tags: Dict[str, List["Token"]]
) -> Optional[List[Tuple[Optional[exprs.Value], Optional[exprs.Value]]]]:
    """Convert an interval specification of a line to a (disjunctive) sequence of
    expressions, if such a specification exists in the given tags.
    """
    if "has_intval_spec" not in tags:
        return None

    ors = tags.get("or", [])
    offsets = cast(List[Optional[int]], [None]) + [tok.offset for tok in ors] + [None]

    boundaries = []
    low_vals = tags.get("lowerbound_value", [])
    low_units = tags.get("lowerbound_unit", [])
    high_vals = tags.get("upperbound_value", [])
    high_units = tags.get("upperbound_unit", [])
    for i in range(0, len(offsets) - 1):
        start_offset, end_offset = offsets[i], offsets[i + 1]
        low_val = get_optional(low_vals, start_offset, end_offset)
        low_unit = get_optional(low_units, start_offset, end_offset)
        high_val = get_optional(high_vals, start_offset, end_offset)
        high_unit = get_optional(high_units, start_offset, end_offset)
        assert low_val or high_val

        low_bound: Optional[exprs.Value]
        high_bound: Optional[exprs.Value]
        if low_val:
            low_bound = exprs.Value(low_val, low_unit)
        else:
            low_bound = None

        if high_val:
            high_bound = exprs.Value(high_val, high_unit)
        else:
            high_bound = None

        boundaries.append((low_bound, high_bound))

    return boundaries


def get_constant_specification_part(start_loc: str, end_2locs: List[str], prefix: str = "") -> str:
    """State machine part implementing the 'constant-specification' rule.

    constant-specification ::=
        "equal" "to" VALUE [ UNIT ]
    """
    assert len(end_2locs) == 2

    locs = {
        "start": start_loc,
        "end1": end_2locs[0],
        "end2": end_2locs[1],
    }
    locs.update(make_loc_names(prefix, "const", 4))

    text = """\
        {const1}; {const2};

        {start}  -> {const1} [EQUAL_KW] tag=has_constant_spec;
        {const1} -> {const2} [TO_KW];
        {const2} -> {end1} [NONCOMMA] tag=const_value;
        {end1}   -> {end2} [NONCOMMA] tag=const_unit;
    """
    return text.format(**locs)


def _convert_constant_spec(tags: Dict[str, List["Token"]]) -> Optional[exprs.Value]:
    """Convert a constant specification of a line to a Value if it exits."""
    if "has_constant_spec" not in tags:
        return None

    val = get_one(tags["const_value"], None, None)
    unit = get_optional(tags.get("const_unit", []), None, None)
    bound = exprs.Value(val, unit)
    return bound


_DEFINE_TYPE_SPEC = """
define_type:
    start initial;
    start -> s1 [DEFINE_KW] tag=define;
    s1 -> s2 [TYPE_KW];

    end accept=define_type;
    s2 -> end [NL_TK];
    s2 -> end [EOF_TK];
"""


def _process_define_type(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    builder.notify_new_section(tags["define"][0], True)


_NEW_TYPE_SPEC = (
    """
new_type:
    start initial;
    end accept=new_type;

    start -> s1 [NAME] tag=new_type;
    s1 -> end [NL_TK];
    s1 -> end [EOF_TK];
"""
    + get_enumeration_specification_part("s1", ["s2", "s3"])
    + """\
    s2 -> end [NL_TK];
    s2 -> end [EOF_TK];
    s3 -> end [NL_TK];
    s3 -> end [EOF_TK];
"""
    + get_unit_specification_part("s1", ["s4"])
    + """\
    s4 -> end [NL_TK];
    s4 -> end [EOF_TK];
"""
    + get_interval_specification_part("s1", ["s5", "s6", "s7", "s8"])
    + """\
    s5 -> end [NL_TK];
    s5 -> end [EOF_TK];
    s6 -> end [NL_TK];
    s6 -> end [EOF_TK];
    s7 -> end [NL_TK];
    s7 -> end [EOF_TK];
    s8 -> end [NL_TK];
    s8 -> end [EOF_TK];
"""
    + get_constant_specification_part("s1", ["s9", "s10"])
    + """\
    s9 -> end [NL_TK];
    s9 -> end [EOF_TK];
    s10 -> end [NL_TK];
    s10 -> end [EOF_TK];
"""
)


def _process_new_type_def(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    new_type = tags["new_type"][0]
    enum_spec = _convert_enum_spec(tags)
    unit_spec = _convert_unit_spec(tags)
    ival_spec = _convert_interval_spec(tags)
    cons_spec = _convert_constant_spec(tags)
    builder.add_typedef(new_type, None, enum_spec, unit_spec, ival_spec, cons_spec)


_TYPE_IS_A_TYPE_SPEC = (
    """
type_is_type:
    start initial;
    end accept=type_is_type;

    start -> s1 [NAME] tag=new_type;
    s1 -> s2 [IS_KW];
    s2 -> s3 [A_KW];
    s3 -> s4 [NAME] tag=old_type;
    s4 -> end [NL_TK];
    s4 -> end [EOF_TK];
"""
    + get_unit_specification_part("s4", ["s5"])
    + """\
    s5 -> end [NL_TK];
    s5 -> end [EOF_TK];
"""
    + get_interval_specification_part("s4", ["s6", "s7", "s8", "s9"])
    + """\
    s6 -> end [NL_TK];
    s6 -> end [EOF_TK];
    s7 -> end [NL_TK];
    s7 -> end [EOF_TK];
    s8 -> end [NL_TK];
    s8 -> end [EOF_TK];
    s9 -> end [NL_TK];
    s9 -> end [EOF_TK];
"""
    + get_constant_specification_part("s4", ["s10", "s11"])
    + """\
    s10 -> end [NL_TK];
    s10 -> end [EOF_TK];
    s11 -> end [NL_TK];
    s11 -> end [EOF_TK];
"""
)


def _process_derived_type_def(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    new_type = tags["new_type"][0]
    exist_type = tags["old_type"][0]
    enum_spec = _convert_enum_spec(tags)
    unit_spec = _convert_unit_spec(tags)
    ival_spec = _convert_interval_spec(tags)
    cons_spec = _convert_constant_spec(tags)
    builder.add_typedef(new_type, exist_type, enum_spec, unit_spec, ival_spec, cons_spec)


_BUNDLE_TYPE_SPEC = """
bundle_type:
    start initial;
    start -> s1 [NAME] tag=bundle_name;
    s1 -> s2 [IS_KW];
    s2 -> s3 [A_KW];
    s3 -> s4 [BUNDLE_KW];
    s4 -> s5 [OF_KW];

    end accept=bundle_type;
    s5 -> end [NL_TK];
    s5 -> end [EOF_TK];
"""


def _process_bundle_type(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    builder.new_bundle_type(tags["bundle_name"][0])


_BUNDLE_FIELD_SPEC = """
bundle_field:
    start initial;
    end accept=bundle_field;

    start -> s1 [STAR_TK];
    s1 -> s2 [NAME] tag=field_name;
    s2 -> s3 [IS_KW];
    s3 -> s4 [A_KW];

    s4 -> s5 [NAME] tag=field_type;

    s5 -> end [NL_TK];
    s5 -> end [EOF_TK];
"""


def _process_bundle_field(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    field_name = tags["field_name"][0]
    field_type = tags["field_type"][0]
    builder.add_bundle_field(field_name, field_type)


MACHINES: typing.MachineTripletList = [
    ("DEFINE_TYPE_MACHINE", _DEFINE_TYPE_SPEC, _process_define_type),
    ("NEW_TYPE_MACHINE", _NEW_TYPE_SPEC, _process_new_type_def),
    ("TYPE_IS_A_TYPE_MACHINE", _TYPE_IS_A_TYPE_SPEC, _process_derived_type_def),
    ("BUNDLE_TYPE_MACHINE", _BUNDLE_TYPE_SPEC, _process_bundle_type),
    ("BUNDLE_FIELD_MACHINE", _BUNDLE_FIELD_SPEC, _process_bundle_field),
]



"""Line matchers for the behavior section."""

from typing import TYPE_CHECKING, Dict, List, Optional, Union, cast

from raesl.compile.ast import exprs
from raesl.compile.machine_files import machine_parts, sub_clause, typing, utils

if TYPE_CHECKING:
    from raesl.compile.scanner import Token
    from raesl.compile.typechecking.ast_builder import AstBuilder


_BEHAVIOR_HEADER_SPEC = """
behavior_header:
    start initial;

    start -> s1 [BEHAVIOR_REQUIREMENT_KW] tag=kind;
    start -> s1 [BEHAVIOR_CONSTRAINT_KW] tag=kind;

    end accept=behavior_header;
    s1 -> end [NL_TK];
"""


def _process_behavior_header(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    behave_kind = tags["kind"][0]
    builder.notify_new_section(behave_kind, False)

    current_comp = builder.compdef_builder.current_component
    assert current_comp is not None
    current_comp.new_behavior_header(behave_kind)


_BEHAVIOR_NAME_SPEC = """
behavior_name:
    start initial;

    start -> s1 [NAME] tag=label;
    s1 -> s2 [COLON_TK];

    end accept=behavior_name;
    s2 -> end [NL_TK];
"""


def _process_new_behavior_function(
    tags: typing.TokensDict, _accept: str, builder: "AstBuilder"
) -> None:
    label = tags["label"][0]

    current_comp = builder.compdef_builder.current_component
    assert current_comp is not None
    current_comp.new_behavior_function(label)


_BEHAVIOR_CASE_SPEC = """
behavior_case:
    start initial;

    start -> s1 [CASE_KW];
    s1 -> s2 [NAME] tag=case_name;
    s2 -> s3 [COLON_TK];

    end accept=behavior_case;
    s3 -> end [NL_TK];
"""


def _process_new_case(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    case_label = tags["case_name"][0]

    current_comp = builder.compdef_builder.current_component
    assert current_comp is not None
    current_comp.behavior_case(case_label)


_BEHAVIOR_WHEN_SPEC = """
behavior_when:
    start initial;

    start -> s1 [WHEN_KW] tag=when;

    end accept=behavior_when;
    s1 -> end [NL_TK];
"""


def _process_normal_when_start(
    tags: typing.TokensDict, _accept: str, builder: "AstBuilder"
) -> None:
    current_comp = builder.compdef_builder.current_component
    assert current_comp is not None
    current_comp.behavior_normal_when(tags["when"][0])


_BEHAVIOR_WHEN_OTHERWISE_SPEC = """
behavior_otherwise:
    start initial;

    start -> s1 [WHEN_KW] tag=when;
    s1 -> s2 [NO_KW];
    s2 -> s3 [OTHER_KW];
    s3 -> s4 [CASE_KW];
    s4 -> s5 [APPLIES_KW];

    end accept=behavior_when;
    s5 -> end [NL_TK];
"""


def _process_otherwise_when_start(
    tags: typing.TokensDict, _accept: str, builder: "AstBuilder"
) -> None:
    current_comp = builder.compdef_builder.current_component
    assert current_comp is not None
    current_comp.behavior_default_when(tags["when"][0])


_BEHAVIOR_WHEN_CONDITION_SPEC = (
    """
when_condition:
    start initial;

    start -> s1 [STAR_TK];
    s1 -> s2 [NAME] tag=condition_name;
    s2 -> s3 [COLON_TK];
    s3 -> s4 [DOTTEDNAME] tag=first_var;
    s4 -> s5 [IS_KW] tag=is_aux;
"""
    + machine_parts.get_compare_op_part("s5", ["s6"], tagname="compare_op")
    + """\
    s6 -> s7 [NONCOMMA] tag=varvalue;
    s7 -> s8 [NONCOMMA] tag=unit;

    s7 -> s3 [OR_KW] tag=or;
    s8 -> s3 [OR_KW] tag=or;

    end accept=behavior_when;
    s7 -> end [NL_TK];
    s8 -> end [NL_TK];
"""
)


def _decode_when_comparisons(
    tags: Dict[str, List["Token"]]
) -> Union[exprs.Disjunction, exprs.RelationComparison]:
    """Decode the expression expressed in the tags from _BEHAVIOR_WHEN_CONDITION_SPEC.

    Much inspired by 'machine_files.decode_disjunctive_comparisons', but this is
    less generic due to lack of 'aux' and objectives.

    Arguments:
        tags: Key pieces of text found in the input.

    Returns:
        The equivalent expression.
    """
    split_offsets = (
        cast(List[Optional[int]], [None]) + [tok.offset for tok in tags.get("or", [])] + [None]
    )
    equations = []

    unit_tags = tags.get("unit", [])
    for index in range(1, len(split_offsets)):
        start_offset = split_offsets[index - 1]
        end_offset = split_offsets[index]

        first_var = utils.get_one(tags["first_var"], start_offset, end_offset)
        lhs = exprs.VariableValue(first_var)

        is_aux = utils.get_one(tags["is_aux"], start_offset, end_offset)
        compare_op = utils.get_one(tags["compare_op"], start_offset, end_offset)
        varvalue = utils.get_one(tags["varvalue"], start_offset, end_offset)
        unit = utils.get_optional(unit_tags, start_offset, end_offset)
        rhs: exprs.DataValue
        if unit is None and sub_clause.guess_is_var(varvalue):
            rhs = exprs.VariableValue(varvalue)
        else:
            rhs = exprs.Value(varvalue, unit)

        comp = exprs.RelationComparison(True, lhs, is_aux, compare_op, rhs)
        equations.append(comp)

    if len(equations) == 1:
        return equations[0]
    else:
        return exprs.Disjunction(equations)


def _process_when_condition(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    name_tok = tags["condition_name"][0]
    expr = _decode_when_comparisons(tags)

    current_comp = builder.compdef_builder.current_component
    assert current_comp is not None
    current_comp.behavior_when_condition(name_tok, expr)


_BEHAVIOR_THEN_SPEC = """
behavior_then:
    start initial;

    start -> s1 [THEN_KW] tag=then;

    end accept=behavior_then;
    s1 -> end [NL_TK];
"""


def _process_then_start(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    current_comp = builder.compdef_builder.current_component
    assert current_comp is not None
    current_comp.behavior_normal_then(tags["then"][0])


_BEHAVIOR_THEN_RESULT_SPEC = (
    """
then_result:
    start initial;

    start -> s1 [STAR_TK];
    s1 -> s2 [NAME] tag=result_name;
    s2 -> s3 [COLON_TK];

    s3 -> s4 [DOTTEDNAME] tag=first_var;
"""
    + machine_parts.get_is_auxiliary_verb_part("s4", ["s5"], tagname="is_aux")
    + """\
    s5 -> s8 [MAXIMIZED_KW] tag=objective;
    s5 -> s8 [MINIMIZED_KW] tag=objective;
"""
    + machine_parts.get_compare_op_part("s5", ["s6"], tagname="compare_op")
    + """\
    s6 -> s7 [NONCOMMA] tag=varvalue;
    s7 -> s8 [NONCOMMA] tag=unit;

    end accept=behavior_then;
    s7 -> end [NL_TK];
    s7 -> end [EOF_TK];
    s8 -> end [NL_TK];
    s8 -> end [EOF_TK];
"""
)


def _decode_result(tags: Dict[str, List["Token"]]) -> exprs.Comparison:
    """Decode a result line from collected tags by matching _BEHAVIOR_THEN_RESULT_SPEC.

    Much inspired by 'machine_files.decode_disjunctive_comparisons', but this is
    simpler as there is no 'or', splitting is thus not needed.

    Arguments:
        tags: Key pieces of text found in the input.

    Returns:
        The expression equivalent to the matched text.
    """
    lhs = exprs.VariableValue(tags["first_var"][0])
    is_aux = tags["is_aux"][0]
    is_constraint = is_aux.tok_text == "is"

    compare_op = tags.get("compare_op")
    if compare_op is not None:
        varvalue = tags["varvalue"][0]
        units = tags.get("unit")
        rhs: exprs.DataValue
        if units is None:
            if sub_clause.guess_is_var(varvalue):
                rhs = exprs.VariableValue(varvalue)
            else:
                rhs = exprs.Value(varvalue, None)
        else:
            rhs = exprs.Value(varvalue, units[0])

        return exprs.RelationComparison(is_constraint, lhs, is_aux, compare_op[0], rhs)

    else:
        objective = tags["objective"][0]
        return exprs.ObjectiveComparison(lhs, is_aux, objective.tok_type == "MAXIMIZED_KW")


def _process_then_result(tags: typing.TokensDict, _accept: str, builder: "AstBuilder") -> None:
    name_tok = tags["result_name"][0]
    expr = _decode_result(tags)

    current_comp = builder.compdef_builder.current_component
    assert current_comp is not None
    current_comp.behavior_then_result(name_tok, expr)


MACHINES: typing.MachineTripletList = [
    ("BEHAVIOR_HEADER_MACHINE", _BEHAVIOR_HEADER_SPEC, _process_behavior_header),
    ("BEHAVIOR_NAME_MACHINE", _BEHAVIOR_NAME_SPEC, _process_new_behavior_function),
    ("BEHAVIOR_CASE_MACHINE", _BEHAVIOR_CASE_SPEC, _process_new_case),
    ("BEHAVIOR_WHEN_MACHINE", _BEHAVIOR_WHEN_SPEC, _process_normal_when_start),
    (
        "BEHAVIOR_WHEN_OTHERWISE_MACHINE",
        _BEHAVIOR_WHEN_OTHERWISE_SPEC,
        _process_otherwise_when_start,
    ),
    (
        "BEHAVIOR_WHEN_CONDITION_MACHINE",
        _BEHAVIOR_WHEN_CONDITION_SPEC,
        _process_when_condition,
    ),
    ("BEHAVIOR_THEN_MACHINE", _BEHAVIOR_THEN_SPEC, _process_then_start),
    ("BEHAVIOR_THEN_RESULT_MACHINE", _BEHAVIOR_THEN_RESULT_SPEC, _process_then_result),
]
